"""
SGP4/SDP4 propagation core translated to JAX.

This module provides the initialization and propagation routines for the
SGP4 (near-Earth) and SDP4 (deep-space) analytical orbit propagators.
The implementation is a direct translation of the reference ``sgp4`` Python
library into JAX-compatible code, supporting JIT compilation and ``vmap``.

The satellite parameters are stored in a flat ``jnp.array`` with named
indices defined in ``_IDX``. The init function (``sgp4_init``) runs at
Python time and returns this array plus a method flag. The propagation
function (``sgp4_propagate``) is a pure JAX function suitable for JIT.
"""

from __future__ import annotations

from collections.abc import Callable
from math import cos as _py_cos
from math import fabs as _py_fabs
from math import pi as _py_pi
from math import sin as _py_sin
from math import sqrt as _py_sqrt
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from astrojax.sgp4._constants import GRAVITY_MODELS, WGS72, EarthGravity
from astrojax.sgp4._tle import parse_omm, parse_tle
from astrojax.sgp4._types import SGP4Elements, elements_to_array

if TYPE_CHECKING:
    from astrojax._gp_record import GPRecord

# ---------------------------------------------------------------------------
# Parameter index layout for the flat params array
# ---------------------------------------------------------------------------
_PARAM_NAMES = [
    # Gravity constants (0-7)
    "radiusearthkm",
    "xke",
    "j2",
    "j3oj2",
    "j4",
    "tumin",
    "mu",
    "j3",
    # TLE-derived orbital elements (8-16)
    "bstar",
    "ecco",
    "argpo",
    "inclo",
    "mo",
    "no_kozai",
    "nodeo",
    "ndot",
    "nddot",
    # Computed by _initl (17-20)
    "no_unkozai",
    "con41",
    "gsto",
    "a",
    # Computed by sgp4init — near-earth (21-45)
    "cc1",
    "cc4",
    "cc5",
    "d2",
    "d3",
    "d4",
    "delmo",
    "eta",
    "argpdot",
    "omgcof",
    "sinmao",
    "t2cof",
    "t3cof",
    "t4cof",
    "t5cof",
    "x1mth2",
    "x7thm1",
    "mdot",
    "nodedot",
    "xlcof",
    "xmcof",
    "nodecf",
    "aycof",
    "isimp",  # 0.0 or 1.0
    # Deep-space parameters (45+)
    "irez",
    "d2201",
    "d2211",
    "d3210",
    "d3222",
    "d4410",
    "d4422",
    "d5220",
    "d5232",
    "d5421",
    "d5433",
    "dedt",
    "del1",
    "del2",
    "del3",
    "didt",
    "dmdt",
    "dnodt",
    "domdt",
    "e3",
    "ee2",
    "peo",
    "pgho",
    "pho",
    "pinco",
    "plo",
    "se2",
    "se3",
    "sgh2",
    "sgh3",
    "sgh4",
    "sh2",
    "sh3",
    "si2",
    "si3",
    "sl2",
    "sl3",
    "sl4",
    "xfact",
    "xgh2",
    "xgh3",
    "xgh4",
    "xh2",
    "xh3",
    "xi2",
    "xi3",
    "xl2",
    "xl3",
    "xl4",
    "xlamo",
    "xli",
    "xni",
    "zmol",
    "zmos",
    "atime",
    # Method flag (96): 0.0 = near-earth, 1.0 = deep-space
    "method",
]

_IDX = {name: i for i, name in enumerate(_PARAM_NAMES)}
_NUM_PARAMS = len(_PARAM_NAMES)

# Shorthand index constants for frequently accessed parameters
_I = _IDX  # alias for brevity in propagation code

# ---------------------------------------------------------------------------
# Python-time helpers (used during init, not under JIT)
# ---------------------------------------------------------------------------
_twopi = 2.0 * _py_pi


def _gstime(jdut1: float) -> float:
    """Compute Greenwich Sidereal Time from Julian date (Python floats)."""
    tut1 = (jdut1 - 2451545.0) / 36525.0
    temp = (
        -6.2e-6 * tut1 * tut1 * tut1
        + 0.093104 * tut1 * tut1
        + (876600.0 * 3600 + 8640184.812866) * tut1
        + 67310.54841
    )
    temp = (temp * (_py_pi / 180.0) / 240.0) % _twopi
    if temp < 0.0:
        temp += _twopi
    return temp


def _initl(
    xke: float,
    j2: float,
    ecco: float,
    epoch: float,
    inclo: float,
    no: float,
    opsmode: str,
) -> tuple:
    """Initialize SGP4 auxiliary quantities (Python floats).

    Args:
        xke: Gravity constant xke.
        j2: J2 zonal harmonic.
        ecco: Eccentricity.
        epoch: Days since 1950 Jan 0.
        inclo: Inclination [rad].
        no: Mean motion (Kozai) [rad/min].
        opsmode: Operation mode ('a' or 'i').

    Returns:
        Tuple of computed quantities.
    """
    x2o3 = 2.0 / 3.0

    eccsq = ecco * ecco
    omeosq = 1.0 - eccsq
    rteosq = _py_sqrt(omeosq)
    cosio = _py_cos(inclo)
    cosio2 = cosio * cosio

    # Un-Kozai the mean motion
    ak = (xke / no) ** x2o3
    d1 = 0.75 * j2 * (3.0 * cosio2 - 1.0) / (rteosq * omeosq)
    del_ = d1 / (ak * ak)
    adel = ak * (1.0 - del_ * del_ - del_ * (1.0 / 3.0 + 134.0 * del_ * del_ / 81.0))
    del_ = d1 / (adel * adel)
    no = no / (1.0 + del_)

    ao = (xke / no) ** x2o3
    sinio = _py_sin(inclo)
    po = ao * omeosq
    con42 = 1.0 - 5.0 * cosio2
    con41 = -con42 - cosio2 - cosio2
    posq = po * po
    rp = ao * (1.0 - ecco)

    # Sidereal time
    if opsmode == "a":
        ts70 = epoch - 7305.0
        ds70 = (ts70 + 1.0e-8) // 1.0
        tfrac = ts70 - ds70
        c1 = 1.72027916940703639e-2
        thgr70 = 1.7321343856509374
        fk5r = 5.07551419432269442e-15
        c1p2p = c1 + _twopi
        gsto = (thgr70 + c1 * ds70 + c1p2p * tfrac + ts70 * ts70 * fk5r) % _twopi
        if gsto < 0.0:
            gsto = gsto + _twopi
    else:
        gsto = _gstime(epoch + 2433281.5)

    return (
        no,
        1.0 / ao,  # ainv
        ao,
        con41,
        con42,
        cosio,
        cosio2,
        eccsq,
        omeosq,
        posq,
        rp,
        rteosq,
        sinio,
        gsto,
    )


# ---------------------------------------------------------------------------
# JAX-compatible init helpers (JIT-safe)
# ---------------------------------------------------------------------------

_twopi_jax = 2.0 * jnp.pi


def _gstime_jax(jdut1: ArrayLike) -> Array:
    """Compute Greenwich Sidereal Time from Julian date (JAX).

    Args:
        jdut1: Julian date (UT1).

    Returns:
        Greenwich sidereal time [rad].
    """
    tut1 = (jdut1 - 2451545.0) / 36525.0
    temp = (
        -6.2e-6 * tut1 * tut1 * tut1
        + 0.093104 * tut1 * tut1
        + (876600.0 * 3600 + 8640184.812866) * tut1
        + 67310.54841
    )
    temp = (temp * (jnp.pi / 180.0) / 240.0) % _twopi_jax
    temp = jnp.where(temp < 0.0, temp + _twopi_jax, temp)
    return temp


def _initl_jax(
    xke: ArrayLike,
    j2: ArrayLike,
    ecco: ArrayLike,
    epoch: ArrayLike,
    inclo: ArrayLike,
    no: ArrayLike,
    opsmode: str,
) -> tuple:
    """Initialize SGP4 auxiliary quantities (JAX, JIT-compatible).

    Args:
        xke: Gravity constant xke.
        j2: J2 zonal harmonic.
        ecco: Eccentricity.
        epoch: Days since 1950 Jan 0.
        inclo: Inclination [rad].
        no: Mean motion (Kozai) [rad/min].
        opsmode: Operation mode ('a' or 'i'). Resolved at trace time.

    Returns:
        Tuple of computed quantities.
    """
    x2o3 = 2.0 / 3.0

    eccsq = ecco * ecco
    omeosq = 1.0 - eccsq
    rteosq = jnp.sqrt(omeosq)
    cosio = jnp.cos(inclo)
    cosio2 = cosio * cosio

    # Un-Kozai the mean motion
    ak = (xke / no) ** x2o3
    d1 = 0.75 * j2 * (3.0 * cosio2 - 1.0) / (rteosq * omeosq)
    del_ = d1 / (ak * ak)
    adel = ak * (1.0 - del_ * del_ - del_ * (1.0 / 3.0 + 134.0 * del_ * del_ / 81.0))
    del_ = d1 / (adel * adel)
    no = no / (1.0 + del_)

    ao = (xke / no) ** x2o3
    sinio = jnp.sin(inclo)
    po = ao * omeosq
    con42 = 1.0 - 5.0 * cosio2
    con41 = -con42 - cosio2 - cosio2
    posq = po * po
    rp = ao * (1.0 - ecco)

    # Sidereal time
    if opsmode == "a":
        ts70 = epoch - 7305.0
        ds70 = jnp.floor(ts70 + 1.0e-8)
        tfrac = ts70 - ds70
        c1 = 1.72027916940703639e-2
        thgr70 = 1.7321343856509374
        fk5r = 5.07551419432269442e-15
        c1p2p = c1 + _twopi_jax
        gsto = (thgr70 + c1 * ds70 + c1p2p * tfrac + ts70 * ts70 * fk5r) % _twopi_jax
        gsto = jnp.where(gsto < 0.0, gsto + _twopi_jax, gsto)
    else:
        gsto = _gstime_jax(epoch + 2433281.5)

    return (
        no,
        1.0 / ao,  # ainv
        ao,
        con41,
        con42,
        cosio,
        cosio2,
        eccsq,
        omeosq,
        posq,
        rp,
        rteosq,
        sinio,
        gsto,
    )


# ---------------------------------------------------------------------------
# SGP4 Initialization (Python-time)
# ---------------------------------------------------------------------------


def sgp4_init(
    elements: SGP4Elements,
    gravity: EarthGravity = WGS72,
    opsmode: str = "i",
) -> tuple[Array, str]:
    """Initialize SGP4 satellite parameters from parsed TLE elements.

    This function runs at Python time (not under JIT). It computes all the
    intermediate parameters needed by the SGP4 propagator and packs them
    into a flat JAX array.

    Args:
        elements: Parsed TLE elements from ``parse_tle`` or ``parse_omm``.
        gravity: Earth gravity model constants.
        opsmode: Operation mode ('i' for improved, 'a' for AFSPC).

    Returns:
        Tuple of ``(params, method)`` where ``params`` is a flat
        ``jnp.array`` of shape ``(_NUM_PARAMS,)`` and ``method`` is
        ``'n'`` (near-Earth) or ``'d'`` (deep-space).
    """
    # Initialize parameter dict with zeros
    d: dict[str, float] = {name: 0.0 for name in _PARAM_NAMES}

    # Store gravity constants
    d["tumin"] = gravity.tumin
    d["mu"] = gravity.mu
    d["radiusearthkm"] = gravity.radiusearthkm
    d["xke"] = gravity.xke
    d["j2"] = gravity.j2
    d["j3"] = gravity.j3
    d["j4"] = gravity.j4
    d["j3oj2"] = gravity.j3oj2

    # Store TLE elements
    d["bstar"] = elements.bstar
    d["ecco"] = elements.ecco
    d["argpo"] = elements.argpo
    d["inclo"] = elements.inclo
    d["mo"] = elements.mo
    d["no_kozai"] = elements.no_kozai
    d["nodeo"] = elements.nodeo
    d["ndot"] = elements.ndot
    d["nddot"] = elements.nddot

    # Compute epoch in days since 1950 Jan 0
    epoch = elements.jdsatepoch + elements.jdsatepochF - 2433281.5

    temp4 = 1.5e-12
    x2o3 = 2.0 / 3.0
    method = "n"

    # Earth constants
    ss = 78.0 / gravity.radiusearthkm + 1.0
    qzms2ttemp = (120.0 - 78.0) / gravity.radiusearthkm
    qzms2t = qzms2ttemp**4
    sfour = ss

    # Initialize auxiliary quantities
    (
        no_unkozai,
        ainv,
        ao,
        con41,
        con42,
        cosio,
        cosio2,
        eccsq,
        omeosq,
        posq,
        rp,
        rteosq,
        sinio,
        gsto,
    ) = _initl(
        gravity.xke,
        gravity.j2,
        elements.ecco,
        epoch,
        elements.inclo,
        elements.no_kozai,
        opsmode,
    )

    d["no_unkozai"] = no_unkozai
    d["con41"] = con41
    d["gsto"] = gsto
    d["a"] = (no_unkozai * gravity.tumin) ** (-2.0 / 3.0)

    if omeosq >= 0.0 or no_unkozai >= 0.0:
        isimp = 0
        if rp < 220.0 / gravity.radiusearthkm + 1.0:
            isimp = 1

        qzms24 = qzms2t
        perige = (rp - 1.0) * gravity.radiusearthkm

        # For perigees below 156 km, s and qoms2t are altered
        if perige < 156.0:
            sfour = perige - 78.0
            if perige < 98.0:
                sfour = 20.0
            qzms24temp = (120.0 - sfour) / gravity.radiusearthkm
            qzms24 = qzms24temp**4
            sfour = sfour / gravity.radiusearthkm + 1.0

        pinvsq = 1.0 / posq
        tsi = 1.0 / (ao - sfour)
        eta = ao * elements.ecco * tsi
        etasq = eta * eta
        eeta = elements.ecco * eta
        psisq = _py_fabs(1.0 - etasq)
        coef = qzms24 * tsi**4
        coef1 = coef / psisq**3.5
        cc2 = (
            coef1
            * no_unkozai
            * (
                ao * (1.0 + 1.5 * etasq + eeta * (4.0 + etasq))
                + 0.375 * gravity.j2 * tsi / psisq * con41 * (8.0 + 3.0 * etasq * (8.0 + etasq))
            )
        )
        cc1 = elements.bstar * cc2
        cc3 = 0.0
        if elements.ecco > 1.0e-4:
            cc3 = -2.0 * coef * tsi * gravity.j3oj2 * no_unkozai * sinio / elements.ecco
        x1mth2 = 1.0 - cosio2
        cc4 = (
            2.0
            * no_unkozai
            * coef1
            * ao
            * omeosq
            * (
                eta * (2.0 + 0.5 * etasq)
                + elements.ecco * (0.5 + 2.0 * etasq)
                - gravity.j2
                * tsi
                / (ao * psisq)
                * (
                    -3.0 * con41 * (1.0 - 2.0 * eeta + etasq * (1.5 - 0.5 * eeta))
                    + 0.75
                    * x1mth2
                    * (2.0 * etasq - eeta * (1.0 + etasq))
                    * _py_cos(2.0 * elements.argpo)
                )
            )
        )
        cc5 = 2.0 * coef1 * ao * omeosq * (1.0 + 2.75 * (etasq + eeta) + eeta * etasq)
        cosio4 = cosio2 * cosio2
        temp1 = 1.5 * gravity.j2 * pinvsq * no_unkozai
        temp2 = 0.5 * temp1 * gravity.j2 * pinvsq
        temp3 = -0.46875 * gravity.j4 * pinvsq * pinvsq * no_unkozai
        mdot = (
            no_unkozai
            + 0.5 * temp1 * rteosq * con41
            + 0.0625 * temp2 * rteosq * (13.0 - 78.0 * cosio2 + 137.0 * cosio4)
        )
        argpdot = (
            -0.5 * temp1 * con42
            + 0.0625 * temp2 * (7.0 - 114.0 * cosio2 + 395.0 * cosio4)
            + temp3 * (3.0 - 36.0 * cosio2 + 49.0 * cosio4)
        )
        xhdot1 = -temp1 * cosio
        nodedot = (
            xhdot1
            + (0.5 * temp2 * (4.0 - 19.0 * cosio2) + 2.0 * temp3 * (3.0 - 7.0 * cosio2)) * cosio
        )
        xpidot = argpdot + nodedot
        omgcof = elements.bstar * cc3 * _py_cos(elements.argpo)
        xmcof = 0.0
        if elements.ecco > 1.0e-4:
            xmcof = -x2o3 * coef * elements.bstar / eeta
        nodecf = 3.5 * omeosq * xhdot1 * cc1
        t2cof = 1.5 * cc1

        # sgp4fix for divide by zero with xinco = 180 deg
        if _py_fabs(cosio + 1.0) > 1.5e-12:
            xlcof = -0.25 * gravity.j3oj2 * sinio * (3.0 + 5.0 * cosio) / (1.0 + cosio)
        else:
            xlcof = -0.25 * gravity.j3oj2 * sinio * (3.0 + 5.0 * cosio) / temp4

        aycof = -0.5 * gravity.j3oj2 * sinio
        delmotemp = 1.0 + eta * _py_cos(elements.mo)
        delmo = delmotemp * delmotemp * delmotemp
        sinmao = _py_sin(elements.mo)
        x7thm1 = 7.0 * cosio2 - 1.0

        # Check for deep space (period >= 225 min)
        if _twopi / no_unkozai >= 225.0:
            method = "d"
            isimp = 1
            d["method"] = 1.0

        # Store near-earth parameters
        d["cc1"] = cc1
        d["cc4"] = cc4
        d["cc5"] = cc5
        d["d2"] = 0.0
        d["d3"] = 0.0
        d["d4"] = 0.0
        d["delmo"] = delmo
        d["eta"] = eta
        d["argpdot"] = argpdot
        d["omgcof"] = omgcof
        d["sinmao"] = sinmao
        d["t2cof"] = t2cof
        d["t3cof"] = 0.0
        d["t4cof"] = 0.0
        d["t5cof"] = 0.0
        d["x1mth2"] = x1mth2
        d["x7thm1"] = x7thm1
        d["mdot"] = mdot
        d["nodedot"] = nodedot
        d["xlcof"] = xlcof
        d["xmcof"] = xmcof
        d["nodecf"] = nodecf
        d["aycof"] = aycof
        d["isimp"] = float(isimp)

        # Deep-space initialization
        if method == "d":
            from astrojax.sgp4._deep_space import deep_space_init

            deep_space_init(
                d,
                elements,
                gravity,
                epoch,
                opsmode,
                no_unkozai,
                ao,
                cosio,
                cosio2,
                eccsq,
                omeosq,
                sinio,
                rteosq,
                con41,
                con42,
                cosio4,
                posq,
                xpidot,
                xhdot1,
                sfour,
                tsi,
                eta,
                etasq,
                eeta,
                psisq,
                cc1,
                cc3,
                coef,
                coef1,
                temp1,
                temp2,
                temp3,
                pinvsq,
            )

        # Higher-order secular terms for non-simplified orbits
        if isimp != 1:
            cc1sq = cc1 * cc1
            d["d2"] = 4.0 * ao * tsi * cc1sq
            temp = d["d2"] * tsi * cc1 / 3.0
            d["d3"] = (17.0 * ao + sfour) * temp
            d["d4"] = 0.5 * temp * ao * tsi * (221.0 * ao + 31.0 * sfour) * cc1
            d["t3cof"] = d["d2"] + 2.0 * cc1sq
            d["t4cof"] = 0.25 * (3.0 * d["d3"] + cc1 * (12.0 * d["d2"] + 10.0 * cc1sq))
            d["t5cof"] = 0.2 * (
                3.0 * d["d4"]
                + 12.0 * cc1 * d["d3"]
                + 6.0 * d["d2"] * d["d2"]
                + 15.0 * cc1sq * (2.0 * d["d2"] + cc1sq)
            )

    # Pack into flat JAX array
    params = jnp.array([d[name] for name in _PARAM_NAMES])

    # Run propagation at t=0 to finalize (matches reference behavior)
    # The reference calls sgp4(satrec, 0.0) here; we do the same
    # but discard the result — it's just to validate init is consistent
    if method == "n":
        _sgp4_propagate_near_earth(params, jnp.float64(0.0))
    else:
        _sgp4_propagate_deep_space(params, jnp.float64(0.0))

    return params, method


# ---------------------------------------------------------------------------
# JIT-compilable SGP4 Initialization
# ---------------------------------------------------------------------------


def sgp4_init_jax(
    elements: Array,
    gravity: EarthGravity = WGS72,
    opsmode: str = "i",
) -> Array:
    """Initialize SGP4 satellite parameters (JAX, JIT-compatible).

    Unlike ``sgp4_init``, this function is compatible with ``jax.jit``,
    ``jax.vmap``, and ``jax.grad``. It accepts a flat array of numeric
    orbital elements and returns a flat parameter array with an embedded
    method flag (0.0 = near-earth, 1.0 = deep-space).

    Use with ``jax.jit``::

        init_jit = jax.jit(sgp4_init_jax, static_argnames=('gravity', 'opsmode'))
        params = init_jit(elements_arr)

    Args:
        elements: Array of shape ``(11,)`` containing the numeric orbital
            elements. Use ``elements_to_array`` to convert from
            ``SGP4Elements``. Order: jdsatepoch, jdsatepochF, no_kozai,
            ecco, inclo, nodeo, argpo, mo, bstar, ndot, nddot.
        gravity: Earth gravity model constants. Treated as a static
            argument under JIT (use ``static_argnames``).
        opsmode: Operation mode ('i' for improved, 'a' for AFSPC).
            Treated as a static argument under JIT.

    Returns:
        Flat ``jnp.array`` of shape ``(_NUM_PARAMS,)`` with all SGP4
        parameters, including a method flag at index ``_IDX["method"]``.
    """
    from astrojax.sgp4._deep_space import _deep_space_init_jax

    # Unpack elements
    jdsatepoch = elements[0]
    jdsatepochF = elements[1]
    no_kozai = elements[2]
    ecco = elements[3]
    inclo = elements[4]
    nodeo = elements[5]
    argpo = elements[6]
    mo = elements[7]
    bstar = elements[8]
    ndot = elements[9]
    nddot = elements[10]

    # Initialize params array
    params = jnp.zeros(_NUM_PARAMS)

    # Store gravity constants (static — known at trace time)
    params = params.at[_IDX["tumin"]].set(gravity.tumin)
    params = params.at[_IDX["mu"]].set(gravity.mu)
    params = params.at[_IDX["radiusearthkm"]].set(gravity.radiusearthkm)
    params = params.at[_IDX["xke"]].set(gravity.xke)
    params = params.at[_IDX["j2"]].set(gravity.j2)
    params = params.at[_IDX["j3"]].set(gravity.j3)
    params = params.at[_IDX["j4"]].set(gravity.j4)
    params = params.at[_IDX["j3oj2"]].set(gravity.j3oj2)

    # Store TLE elements
    params = params.at[_IDX["bstar"]].set(bstar)
    params = params.at[_IDX["ecco"]].set(ecco)
    params = params.at[_IDX["argpo"]].set(argpo)
    params = params.at[_IDX["inclo"]].set(inclo)
    params = params.at[_IDX["mo"]].set(mo)
    params = params.at[_IDX["no_kozai"]].set(no_kozai)
    params = params.at[_IDX["nodeo"]].set(nodeo)
    params = params.at[_IDX["ndot"]].set(ndot)
    params = params.at[_IDX["nddot"]].set(nddot)

    # Compute epoch in days since 1950 Jan 0
    epoch = jdsatepoch + jdsatepochF - 2433281.5

    temp4 = 1.5e-12
    x2o3 = 2.0 / 3.0
    twopi = 2.0 * jnp.pi

    # Earth constants
    ss = 78.0 / gravity.radiusearthkm + 1.0
    qzms2ttemp = (120.0 - 78.0) / gravity.radiusearthkm
    qzms2t = qzms2ttemp**4
    sfour_base = ss

    # Initialize auxiliary quantities
    (
        no_unkozai, ainv, ao, con41, con42, cosio, cosio2,
        eccsq, omeosq, posq, rp, rteosq, sinio, gsto,
    ) = _initl_jax(gravity.xke, gravity.j2, ecco, epoch, inclo, no_kozai, opsmode)

    params = params.at[_IDX["no_unkozai"]].set(no_unkozai)
    params = params.at[_IDX["con41"]].set(con41)
    params = params.at[_IDX["gsto"]].set(gsto)
    params = params.at[_IDX["a"]].set((no_unkozai * gravity.tumin) ** (-2.0 / 3.0))

    # Perigee-based s/qoms2t adjustments
    isimp = jnp.where(rp < 220.0 / gravity.radiusearthkm + 1.0, 1.0, 0.0)

    qzms24 = qzms2t
    perige = (rp - 1.0) * gravity.radiusearthkm

    # For perigees below 156 km
    sfour_low = jnp.where(perige < 98.0, 20.0, perige - 78.0)
    qzms24temp_low = (120.0 - sfour_low) / gravity.radiusearthkm
    qzms24_low = qzms24temp_low**4
    sfour_low_adj = sfour_low / gravity.radiusearthkm + 1.0

    is_low_perige = perige < 156.0
    sfour = jnp.where(is_low_perige, sfour_low_adj, sfour_base)
    qzms24 = jnp.where(is_low_perige, qzms24_low, qzms24)

    pinvsq = 1.0 / posq
    tsi = 1.0 / (ao - sfour)
    eta = ao * ecco * tsi
    etasq = eta * eta
    eeta = ecco * eta
    psisq = jnp.abs(1.0 - etasq)
    coef = qzms24 * tsi**4
    coef1 = coef / psisq**3.5
    cc2 = (
        coef1
        * no_unkozai
        * (
            ao * (1.0 + 1.5 * etasq + eeta * (4.0 + etasq))
            + 0.375 * gravity.j2 * tsi / psisq * con41 * (8.0 + 3.0 * etasq * (8.0 + etasq))
        )
    )
    cc1 = bstar * cc2
    cc3 = jnp.where(
        ecco > 1.0e-4,
        -2.0 * coef * tsi * gravity.j3oj2 * no_unkozai * sinio / ecco,
        0.0,
    )
    x1mth2 = 1.0 - cosio2
    cc4 = (
        2.0
        * no_unkozai
        * coef1
        * ao
        * omeosq
        * (
            eta * (2.0 + 0.5 * etasq)
            + ecco * (0.5 + 2.0 * etasq)
            - gravity.j2
            * tsi
            / (ao * psisq)
            * (
                -3.0 * con41 * (1.0 - 2.0 * eeta + etasq * (1.5 - 0.5 * eeta))
                + 0.75
                * x1mth2
                * (2.0 * etasq - eeta * (1.0 + etasq))
                * jnp.cos(2.0 * argpo)
            )
        )
    )
    cc5 = 2.0 * coef1 * ao * omeosq * (1.0 + 2.75 * (etasq + eeta) + eeta * etasq)
    cosio4 = cosio2 * cosio2
    temp1 = 1.5 * gravity.j2 * pinvsq * no_unkozai
    temp2 = 0.5 * temp1 * gravity.j2 * pinvsq
    temp3 = -0.46875 * gravity.j4 * pinvsq * pinvsq * no_unkozai
    mdot = (
        no_unkozai
        + 0.5 * temp1 * rteosq * con41
        + 0.0625 * temp2 * rteosq * (13.0 - 78.0 * cosio2 + 137.0 * cosio4)
    )
    argpdot = (
        -0.5 * temp1 * con42
        + 0.0625 * temp2 * (7.0 - 114.0 * cosio2 + 395.0 * cosio4)
        + temp3 * (3.0 - 36.0 * cosio2 + 49.0 * cosio4)
    )
    xhdot1 = -temp1 * cosio
    nodedot = (
        xhdot1
        + (0.5 * temp2 * (4.0 - 19.0 * cosio2) + 2.0 * temp3 * (3.0 - 7.0 * cosio2)) * cosio
    )
    xpidot = argpdot + nodedot
    omgcof = bstar * cc3 * jnp.cos(argpo)
    xmcof = jnp.where(ecco > 1.0e-4, -x2o3 * coef * bstar / eeta, 0.0)
    nodecf = 3.5 * omeosq * xhdot1 * cc1
    t2cof = 1.5 * cc1

    # Cosio singularity fix
    xlcof = jnp.where(
        jnp.abs(cosio + 1.0) > 1.5e-12,
        -0.25 * gravity.j3oj2 * sinio * (3.0 + 5.0 * cosio) / (1.0 + cosio),
        -0.25 * gravity.j3oj2 * sinio * (3.0 + 5.0 * cosio) / temp4,
    )

    aycof = -0.5 * gravity.j3oj2 * sinio
    delmotemp = 1.0 + eta * jnp.cos(mo)
    delmo = delmotemp**3
    sinmao = jnp.sin(mo)
    x7thm1 = 7.0 * cosio2 - 1.0

    # Deep-space detection
    is_deep_space = twopi / no_unkozai >= 225.0
    isimp = jnp.where(is_deep_space, 1.0, isimp)

    # Store near-earth parameters
    params = params.at[_IDX["cc1"]].set(cc1)
    params = params.at[_IDX["cc4"]].set(cc4)
    params = params.at[_IDX["cc5"]].set(cc5)
    params = params.at[_IDX["delmo"]].set(delmo)
    params = params.at[_IDX["eta"]].set(eta)
    params = params.at[_IDX["argpdot"]].set(argpdot)
    params = params.at[_IDX["omgcof"]].set(omgcof)
    params = params.at[_IDX["sinmao"]].set(sinmao)
    params = params.at[_IDX["t2cof"]].set(t2cof)
    params = params.at[_IDX["x1mth2"]].set(x1mth2)
    params = params.at[_IDX["x7thm1"]].set(x7thm1)
    params = params.at[_IDX["mdot"]].set(mdot)
    params = params.at[_IDX["nodedot"]].set(nodedot)
    params = params.at[_IDX["xlcof"]].set(xlcof)
    params = params.at[_IDX["xmcof"]].set(xmcof)
    params = params.at[_IDX["nodecf"]].set(nodecf)
    params = params.at[_IDX["aycof"]].set(aycof)
    params = params.at[_IDX["isimp"]].set(isimp)

    # Deep-space initialization (always computed, conditionally applied)
    ds_values = _deep_space_init_jax(
        ecco=ecco, argpo=argpo, inclo=inclo, mo=mo, nodeo=nodeo,
        no_unkozai=no_unkozai, epoch=epoch, gsto=gsto, mdot=mdot,
        nodedot=nodedot, xpidot=xpidot, eccsq=eccsq, xke=gravity.xke,
    )
    for name, val in ds_values.items():
        params = params.at[_IDX[name]].set(
            jnp.where(is_deep_space, val, params[_IDX[name]])
        )

    # Higher-order secular terms for non-simplified orbits
    is_not_simple = isimp < 0.5
    cc1sq = cc1 * cc1
    d2_val = 4.0 * ao * tsi * cc1sq
    temp_d = d2_val * tsi * cc1 / 3.0
    d3_val = (17.0 * ao + sfour) * temp_d
    d4_val = 0.5 * temp_d * ao * tsi * (221.0 * ao + 31.0 * sfour) * cc1
    t3cof_val = d2_val + 2.0 * cc1sq
    t4cof_val = 0.25 * (3.0 * d3_val + cc1 * (12.0 * d2_val + 10.0 * cc1sq))
    t5cof_val = 0.2 * (
        3.0 * d4_val
        + 12.0 * cc1 * d3_val
        + 6.0 * d2_val * d2_val
        + 15.0 * cc1sq * (2.0 * d2_val + cc1sq)
    )

    params = params.at[_IDX["d2"]].set(jnp.where(is_not_simple, d2_val, 0.0))
    params = params.at[_IDX["d3"]].set(jnp.where(is_not_simple, d3_val, 0.0))
    params = params.at[_IDX["d4"]].set(jnp.where(is_not_simple, d4_val, 0.0))
    params = params.at[_IDX["t3cof"]].set(jnp.where(is_not_simple, t3cof_val, 0.0))
    params = params.at[_IDX["t4cof"]].set(jnp.where(is_not_simple, t4cof_val, 0.0))
    params = params.at[_IDX["t5cof"]].set(jnp.where(is_not_simple, t5cof_val, 0.0))

    # Method flag
    params = params.at[_IDX["method"]].set(jnp.where(is_deep_space, 1.0, 0.0))

    return params


# ---------------------------------------------------------------------------
# SGP4 Propagation (JAX — JIT-compatible)
# ---------------------------------------------------------------------------


def _sgp4_propagate_near_earth(params: Array, tsince: ArrayLike) -> tuple[Array, Array]:
    """Propagate near-Earth satellite using SGP4 (JAX, JIT-compatible).

    Args:
        params: Flat parameter array from ``sgp4_init``.
        tsince: Time since epoch in minutes.

    Returns:
        Tuple of ``(r, v)`` where ``r`` is position [km] and ``v`` is
        velocity [km/s], both as 3-element arrays in the TEME frame.
        Returns NaN arrays if the satellite has decayed or an error occurs.
    """
    twopi = 2.0 * jnp.pi
    x2o3 = 2.0 / 3.0

    # Unpack parameters
    p = params
    radiusearthkm = p[_I["radiusearthkm"]]
    xke = p[_I["xke"]]
    j2 = p[_I["j2"]]
    bstar = p[_I["bstar"]]
    ecco = p[_I["ecco"]]
    argpo = p[_I["argpo"]]
    inclo = p[_I["inclo"]]
    mo = p[_I["mo"]]
    nodeo = p[_I["nodeo"]]
    no_unkozai = p[_I["no_unkozai"]]
    con41 = p[_I["con41"]]
    cc1 = p[_I["cc1"]]
    cc4 = p[_I["cc4"]]
    cc5 = p[_I["cc5"]]
    d2 = p[_I["d2"]]
    d3 = p[_I["d3"]]
    d4 = p[_I["d4"]]
    delmo = p[_I["delmo"]]
    eta = p[_I["eta"]]
    argpdot = p[_I["argpdot"]]
    omgcof = p[_I["omgcof"]]
    sinmao = p[_I["sinmao"]]
    t2cof = p[_I["t2cof"]]
    t3cof = p[_I["t3cof"]]
    t4cof = p[_I["t4cof"]]
    t5cof = p[_I["t5cof"]]
    x1mth2 = p[_I["x1mth2"]]
    x7thm1 = p[_I["x7thm1"]]
    mdot = p[_I["mdot"]]
    nodedot = p[_I["nodedot"]]
    xlcof = p[_I["xlcof"]]
    xmcof = p[_I["xmcof"]]
    nodecf = p[_I["nodecf"]]
    aycof = p[_I["aycof"]]
    isimp = p[_I["isimp"]]

    vkmpersec = radiusearthkm * xke / 60.0

    # --- Update for secular gravity and atmospheric drag ---
    t = tsince
    xmdf = mo + mdot * t
    argpdf = argpo + argpdot * t
    nodedf = nodeo + nodedot * t
    argpm = argpdf
    mm = xmdf
    t2 = t * t
    nodem = nodedf + nodecf * t2
    tempa = 1.0 - cc1 * t
    tempe = bstar * cc4 * t
    templ = t2cof * t2

    # Non-simplified orbit corrections
    is_not_simple = isimp < 0.5  # isimp == 0
    delomg = omgcof * t
    delmtemp = 1.0 + eta * jnp.cos(xmdf)
    delm = xmcof * (delmtemp * delmtemp * delmtemp - delmo)
    temp_corr = delomg + delm
    mm_ns = xmdf + temp_corr
    argpm_ns = argpdf - temp_corr
    t3 = t2 * t
    t4 = t3 * t
    tempa_ns = tempa - d2 * t2 - d3 * t3 - d4 * t4
    tempe_ns = tempe + bstar * cc5 * (jnp.sin(mm_ns) - sinmao)
    templ_ns = templ + t3cof * t3 + t4 * (t4cof + t * t5cof)

    mm = jnp.where(is_not_simple, mm_ns, mm)
    argpm = jnp.where(is_not_simple, argpm_ns, argpm)
    tempa = jnp.where(is_not_simple, tempa_ns, tempa)
    tempe = jnp.where(is_not_simple, tempe_ns, tempe)
    templ = jnp.where(is_not_simple, templ_ns, templ)

    nm = no_unkozai
    em = ecco
    inclm = inclo

    # Error check: nm <= 0
    nm_ok = nm > 0.0

    am = (xke / nm) ** x2o3 * tempa * tempa
    nm = xke / am**1.5
    em = em - tempe

    # Eccentricity bounds check
    em_ok = (em < 1.0) & (em >= -0.001)
    em = jnp.clip(em, 1.0e-6, 0.999999)

    mm = mm + no_unkozai * templ
    xlm = mm + argpm + nodem

    nodem = nodem % twopi
    argpm = argpm % twopi
    xlm = xlm % twopi
    mm = (xlm - argpm - nodem) % twopi

    # --- Compute extra mean quantities ---
    sinim = jnp.sin(inclm)
    cosim = jnp.cos(inclm)

    # Near-Earth: no lunar-solar periodics
    ep = em
    xincp = inclm
    argpp = argpm
    nodep = nodem
    mp = mm
    sinip = sinim
    cosip = cosim

    # --- Long period periodics ---
    axnl = ep * jnp.cos(argpp)
    temp_lp = 1.0 / (am * (1.0 - ep * ep))
    aynl = ep * jnp.sin(argpp) + temp_lp * aycof
    xl = mp + argpp + nodep + temp_lp * xlcof * axnl

    # --- Solve Kepler's equation ---
    u = (xl - nodep) % twopi

    def kepler_step(i, eo1):
        sineo1 = jnp.sin(eo1)
        coseo1 = jnp.cos(eo1)
        denom = 1.0 - coseo1 * axnl - sineo1 * aynl
        tem5 = (u - aynl * coseo1 + axnl * sineo1 - eo1) / denom
        tem5 = jnp.clip(tem5, -0.95, 0.95)
        return eo1 + tem5

    eo1 = jax.lax.fori_loop(0, 10, kepler_step, u)

    sineo1 = jnp.sin(eo1)
    coseo1 = jnp.cos(eo1)

    # --- Short period preliminary quantities ---
    ecose = axnl * coseo1 + aynl * sineo1
    esine = axnl * sineo1 - aynl * coseo1
    el2 = axnl * axnl + aynl * aynl
    pl = am * (1.0 - el2)

    # Error check: pl < 0
    pl_ok = pl >= 0.0

    rl = am * (1.0 - ecose)
    rdotl = jnp.sqrt(am) * esine / rl
    rvdotl = jnp.sqrt(pl) / rl
    betal = jnp.sqrt(1.0 - el2)
    temp_sp = esine / (1.0 + betal)
    sinu = am / rl * (sineo1 - aynl - axnl * temp_sp)
    cosu = am / rl * (coseo1 - axnl + aynl * temp_sp)
    su = jnp.arctan2(sinu, cosu)
    sin2u = (cosu + cosu) * sinu
    cos2u = 1.0 - 2.0 * sinu * sinu
    temp_sp2 = 1.0 / pl
    temp1 = 0.5 * j2 * temp_sp2
    temp2 = temp1 * temp_sp2

    # --- Update for short period periodics ---
    mrt = rl * (1.0 - 1.5 * temp2 * betal * con41) + 0.5 * temp1 * x1mth2 * cos2u
    su = su - 0.25 * temp2 * x7thm1 * sin2u
    xnode = nodep + 1.5 * temp2 * cosip * sin2u
    xinc = xincp + 1.5 * temp2 * cosip * sinip * cos2u
    mvt = rdotl - nm * temp1 * x1mth2 * sin2u / xke
    rvdot = rvdotl + nm * temp1 * (x1mth2 * cos2u + 1.5 * con41) / xke

    # --- Orientation vectors ---
    sinsu = jnp.sin(su)
    cossu = jnp.cos(su)
    snod = jnp.sin(xnode)
    cnod = jnp.cos(xnode)
    sini = jnp.sin(xinc)
    cosi = jnp.cos(xinc)
    xmx = -snod * cosi
    xmy = cnod * cosi
    ux = xmx * sinsu + cnod * cossu
    uy = xmy * sinsu + snod * cossu
    uz = sini * sinsu
    vx = xmx * cossu - cnod * sinsu
    vy = xmy * cossu - snod * sinsu
    vz = sini * cossu

    # --- Position and velocity (km and km/s) ---
    _mr = mrt * radiusearthkm
    r = jnp.array([_mr * ux, _mr * uy, _mr * uz])
    v = jnp.array(
        [
            (mvt * ux + rvdot * vx) * vkmpersec,
            (mvt * uy + rvdot * vy) * vkmpersec,
            (mvt * uz + rvdot * vz) * vkmpersec,
        ]
    )

    # Mask errors with NaN
    valid = nm_ok & em_ok & pl_ok & (mrt >= 1.0)
    nan3 = jnp.full(3, jnp.nan)
    r = jnp.where(valid, r, nan3)
    v = jnp.where(valid, v, nan3)

    return r, v


def _sgp4_propagate_deep_space(params: Array, tsince: ArrayLike) -> tuple[Array, Array]:
    """Propagate deep-space satellite using SDP4 (JAX, JIT-compatible).

    Args:
        params: Flat parameter array from ``sgp4_init``.
        tsince: Time since epoch in minutes.

    Returns:
        Tuple of ``(r, v)`` where ``r`` is position [km] and ``v`` is
        velocity [km/s], both as 3-element arrays in the TEME frame.
    """
    from astrojax.sgp4._deep_space import sgp4_propagate_deep_space_impl

    return sgp4_propagate_deep_space_impl(params, tsince, _I)


def sgp4_propagate(params: Array, tsince: ArrayLike, method: str) -> tuple[Array, Array]:
    """Propagate a satellite using SGP4/SDP4.

    This is the main propagation entry point. The ``method`` flag selects
    near-Earth ('n') or deep-space ('d') code path at Python trace time,
    making it compatible with ``jax.jit``.

    Args:
        params: Flat parameter array from ``sgp4_init``.
        tsince: Time since epoch in minutes.
        method: ``'n'`` for near-Earth SGP4, ``'d'`` for deep-space SDP4.

    Returns:
        Tuple of ``(r, v)`` where ``r`` is position [km] and ``v`` is
        velocity [km/s], both as 3-element arrays in the TEME frame.
    """
    if method == "n":
        return _sgp4_propagate_near_earth(params, tsince)
    else:
        return _sgp4_propagate_deep_space(params, tsince)


def sgp4_propagate_unified(params: Array, tsince: ArrayLike) -> tuple[Array, Array]:
    """Propagate a satellite using SGP4/SDP4 with auto-detection (JAX, JIT-compatible).

    Unlike ``sgp4_propagate``, this function does not require a separate
    ``method`` string. It reads the method flag from ``params`` (set by
    ``sgp4_init_jax`` or ``sgp4_init``) and dispatches to the correct
    propagation branch using ``jax.lax.cond``.

    This enables ``vmap`` over satellites with mixed near-earth and
    deep-space orbits in a single batch.

    Args:
        params: Flat parameter array from ``sgp4_init`` or ``sgp4_init_jax``.
        tsince: Time since epoch in minutes.

    Returns:
        Tuple of ``(r, v)`` where ``r`` is position [km] and ``v`` is
        velocity [km/s], both as 3-element arrays in the TEME frame.
    """
    is_deep = params[_IDX["method"]] > 0.5
    return jax.lax.cond(
        is_deep,
        _sgp4_propagate_deep_space,
        _sgp4_propagate_near_earth,
        params,
        tsince,
    )


def _create_propagator_from_elements(
    elements: SGP4Elements,
    gravity: EarthGravity,
) -> tuple[Array, Callable[[ArrayLike], tuple[Array, Array]]]:
    """Shared helper: init SGP4 from elements and return (params, propagate_fn).

    Args:
        elements: Parsed orbital elements.
        gravity: Earth gravity model constants.

    Returns:
        Tuple of ``(params, propagate_fn)`` where:
            - ``params`` is a flat ``jnp.array`` of satellite parameters
            - ``propagate_fn(tsince)`` takes time since epoch in minutes
              and returns ``(r_km, v_kms)`` in the TEME frame
    """
    params, method = sgp4_init(elements, gravity)

    if method == "n":
        propagate_fn = lambda tsince: _sgp4_propagate_near_earth(params, tsince)  # noqa: E731
    else:
        propagate_fn = lambda tsince: _sgp4_propagate_deep_space(params, tsince)  # noqa: E731

    return params, propagate_fn


def create_sgp4_propagator(
    line1: str,
    line2: str,
    gravity: str | EarthGravity = "wgs72",
) -> tuple[Array, Callable[[ArrayLike], tuple[Array, Array]]]:
    """Create a JIT-compatible SGP4 propagator from TLE lines.

    This is the functional API that returns a parameter array and a
    propagation closure suitable for ``jax.jit`` and ``jax.vmap``.

    Args:
        line1: First TLE line.
        line2: Second TLE line.
        gravity: Gravity model name (``'wgs72'``, ``'wgs84'``, ``'wgs72old'``)
            or an ``EarthGravity`` instance.

    Returns:
        Tuple of ``(params, propagate_fn)`` where:
            - ``params`` is a flat ``jnp.array`` of satellite parameters
            - ``propagate_fn(tsince)`` takes time since epoch in minutes
              and returns ``(r_km, v_kms)`` in the TEME frame
    """
    elements = parse_tle(line1, line2)

    if isinstance(gravity, str):
        gravity = GRAVITY_MODELS[gravity.lower()]

    return _create_propagator_from_elements(elements, gravity)


def create_sgp4_propagator_from_elements(
    elements: SGP4Elements,
    gravity: str | EarthGravity = "wgs72",
) -> tuple[Array, Callable[[ArrayLike], tuple[Array, Array]]]:
    """Create a JIT-compatible SGP4 propagator from pre-parsed elements.

    Accepts an :class:`SGP4Elements` dataclass (e.g. from ``parse_tle``
    or ``parse_omm``) and returns a parameter array and propagation closure.

    Args:
        elements: Parsed orbital elements from ``parse_tle`` or ``parse_omm``.
        gravity: Gravity model name (``'wgs72'``, ``'wgs84'``, ``'wgs72old'``)
            or an ``EarthGravity`` instance.

    Returns:
        Tuple of ``(params, propagate_fn)`` where:
            - ``params`` is a flat ``jnp.array`` of satellite parameters
            - ``propagate_fn(tsince)`` takes time since epoch in minutes
              and returns ``(r_km, v_kms)`` in the TEME frame
    """
    if isinstance(gravity, str):
        gravity = GRAVITY_MODELS[gravity.lower()]

    return _create_propagator_from_elements(elements, gravity)


def create_sgp4_propagator_from_omm(
    fields: dict[str, str],
    gravity: str | EarthGravity = "wgs72",
) -> tuple[Array, Callable[[ArrayLike], tuple[Array, Array]]]:
    """Create a JIT-compatible SGP4 propagator from OMM fields.

    Accepts a dictionary of OMM (Orbit Mean-Elements Message) field names
    and string values, as would be parsed from a CSV or XML OMM record.

    Args:
        fields: Dictionary mapping OMM field names to string values.
            See :func:`~astrojax.sgp4.parse_omm` for required keys.
        gravity: Gravity model name (``'wgs72'``, ``'wgs84'``, ``'wgs72old'``)
            or an ``EarthGravity`` instance.

    Returns:
        Tuple of ``(params, propagate_fn)`` where:
            - ``params`` is a flat ``jnp.array`` of satellite parameters
            - ``propagate_fn(tsince)`` takes time since epoch in minutes
              and returns ``(r_km, v_kms)`` in the TEME frame

    Raises:
        KeyError: If a required OMM field is missing.
    """
    elements = parse_omm(fields)

    if isinstance(gravity, str):
        gravity = GRAVITY_MODELS[gravity.lower()]

    return _create_propagator_from_elements(elements, gravity)


def create_sgp4_propagator_from_gp_record(
    record: GPRecord,
    gravity: str | EarthGravity = "wgs72",
) -> tuple[Array, Callable[[ArrayLike], tuple[Array, Array]]]:
    """Create a JIT-compatible SGP4 propagator from a GPRecord.

    Accepts a :class:`~astrojax.GPRecord` (e.g. from CelesTrak or
    SpaceTrack) and returns a parameter array and propagation closure.

    Args:
        record: A GP record from CelesTrak or SpaceTrack.
        gravity: Gravity model name (``'wgs72'``, ``'wgs84'``, ``'wgs72old'``)
            or an ``EarthGravity`` instance.

    Returns:
        Tuple of ``(params, propagate_fn)`` where:
            - ``params`` is a flat ``jnp.array`` of satellite parameters
            - ``propagate_fn(tsince)`` takes time since epoch in minutes
              and returns ``(r_km, v_kms)`` in the TEME frame

    Raises:
        KeyError: If the record is missing required fields.
    """
    elements = record.to_sgp4_elements()

    if isinstance(gravity, str):
        gravity = GRAVITY_MODELS[gravity.lower()]

    return _create_propagator_from_elements(elements, gravity)


# ---------------------------------------------------------------------------
# Convenience converters for JIT-compilable init
# ---------------------------------------------------------------------------


def omm_to_array(fields: dict[str, str]) -> Array:
    """Convert OMM fields dict to a flat JAX array for use with ``sgp4_init_jax``.

    This is a convenience wrapper that calls ``parse_omm`` followed by
    ``elements_to_array``.

    Args:
        fields: Dictionary mapping OMM field names to string values.
            See :func:`~astrojax.sgp4.parse_omm` for required keys.

    Returns:
        Array of shape ``(11,)`` containing the numeric orbital elements.

    Raises:
        KeyError: If a required OMM field is missing.
    """
    return elements_to_array(parse_omm(fields))


def gp_record_to_array(record: GPRecord) -> Array:
    """Convert a GPRecord to a flat JAX array for use with ``sgp4_init_jax``.

    This is a convenience wrapper that calls ``record.to_sgp4_elements()``
    followed by ``elements_to_array``.

    Args:
        record: A GP record from CelesTrak or SpaceTrack.

    Returns:
        Array of shape ``(11,)`` containing the numeric orbital elements.

    Raises:
        KeyError: If the record is missing required fields.
    """
    return elements_to_array(record.to_sgp4_elements())



"""Earth Orientation Parameters (EOP) for JAX-compatible lookups.

Provides JIT-compatible EOP data storage and interpolation using sorted
JAX arrays and ``jnp.searchsorted``. All query functions work inside
``jax.jit``, ``jax.vmap``, and ``jax.grad``.

Typical usage::

    from astrojax.eop import load_default_eop, get_ut1_utc
    eop = load_default_eop()
    ut1_utc = get_ut1_utc(eop, 59569.5)
"""

from astrojax.eop._download import download_standard_eop_file
from astrojax.eop._lookup import get_dxdy, get_eop, get_lod, get_pm, get_ut1_utc
from astrojax.eop._providers import (
    load_cached_eop,
    load_default_eop,
    load_eop_from_file,
    load_eop_from_standard_file,
    static_eop,
    zero_eop,
)
from astrojax.eop._types import EOPData, EOPExtrapolation

__all__ = [
    "EOPData",
    "EOPExtrapolation",
    "download_standard_eop_file",
    "get_dxdy",
    "get_eop",
    "get_lod",
    "get_pm",
    "get_ut1_utc",
    "load_cached_eop",
    "load_default_eop",
    "load_eop_from_file",
    "load_eop_from_standard_file",
    "static_eop",
    "zero_eop",
]



"""Factory functions for creating SpaceWeatherData instances.

Provides convenience constructors for common space weather configurations:

- :func:`static_space_weather`: Constant space weather values (useful for
  testing or when specific values are known).
- :func:`zero_space_weather`: All-zero space weather.
- :func:`load_sw_from_file`: Load from CSSI format file.
- :func:`load_default_sw`: Load bundled ``sw19571001.txt`` data.
- :func:`load_cached_sw`: Load from a local cache, downloading fresh data
  from CelesTrak when stale.
"""

from __future__ import annotations

import importlib.resources
import logging
from pathlib import Path

import jax.numpy as jnp

from astrojax.config import get_dtype
from astrojax.space_weather._download import _SW_FILENAME, download_sw_file
from astrojax.space_weather._parsers import parse_cssi_file
from astrojax.space_weather._types import SpaceWeatherData
from astrojax.utils.caching import get_sw_cache_dir, is_file_stale

logger = logging.getLogger(__name__)

_DEFAULT_MAX_AGE_DAYS: float = 7.0
"""Default maximum age for cached space weather data in days."""


def static_space_weather(
    ap: float = 4.0,
    f107: float = 150.0,
    f107a: float = 150.0,
    kp: float = 1.0,
    mjd_min: float = 0.0,
    mjd_max: float = 99999.0,
) -> SpaceWeatherData:
    """Create a SpaceWeatherData with constant values.

    Useful for testing or when specific constant space weather values are
    known. The resulting dataset contains two points (at mjd_min and mjd_max)
    with identical values, so lookups return the constant everywhere.

    Args:
        ap: Constant Ap index. Default: 4.0.
        f107: Constant F10.7 flux [sfu]. Default: 150.0.
        f107a: Constant 81-day average F10.7 flux [sfu]. Default: 150.0.
        kp: Constant Kp index. Default: 1.0.
        mjd_min: Start of the valid MJD range. Default: 0.0.
        mjd_max: End of the valid MJD range. Default: 99999.0.

    Returns:
        SpaceWeatherData with constant values.

    Examples:
        ```python
        from astrojax.space_weather import static_space_weather, get_sw_f107_obs
        sw = static_space_weather(f107=200.0)
        val = get_sw_f107_obs(sw, 59569.0)  # returns ~200.0
        ```
    """
    dtype = get_dtype()
    mjd_arr = jnp.array([mjd_min, mjd_max], dtype=dtype)
    kp_arr = jnp.full((2, 8), kp, dtype=dtype)
    ap_arr = jnp.full((2, 8), ap, dtype=dtype)
    return SpaceWeatherData(
        mjd=mjd_arr,
        kp=kp_arr,
        ap=ap_arr,
        ap_daily=jnp.array([ap, ap], dtype=dtype),
        f107_obs=jnp.array([f107, f107], dtype=dtype),
        f107_adj=jnp.array([f107, f107], dtype=dtype),
        f107_obs_ctr81=jnp.array([f107a, f107a], dtype=dtype),
        f107_obs_lst81=jnp.array([f107a, f107a], dtype=dtype),
        f107_adj_ctr81=jnp.array([f107a, f107a], dtype=dtype),
        f107_adj_lst81=jnp.array([f107a, f107a], dtype=dtype),
        mjd_min=jnp.array(mjd_min, dtype=dtype),
        mjd_max=jnp.array(mjd_max, dtype=dtype),
    )


def zero_space_weather() -> SpaceWeatherData:
    """Create a SpaceWeatherData with all-zero values.

    Returns:
        SpaceWeatherData with all values set to zero.

    Examples:
        ```python
        from astrojax.space_weather import zero_space_weather, get_sw_f107_obs
        sw = zero_space_weather()
        val = get_sw_f107_obs(sw, 59569.0)  # returns 0.0
        ```
    """
    return static_space_weather(ap=0.0, f107=0.0, f107a=0.0, kp=0.0)


def load_sw_from_file(filepath: str | Path) -> SpaceWeatherData:
    """Load space weather data from a CSSI format file.

    Parses the file, converts lists to JAX arrays, and computes metadata
    scalars (mjd_min, mjd_max).

    Args:
        filepath: Path to a CSSI space weather file
            (e.g. ``sw19571001.txt``).

    Returns:
        SpaceWeatherData ready for JIT-compatible lookups.

    Raises:
        FileNotFoundError: If the file does not exist.
        ValueError: If no valid data is found.

    Examples:
        ```python
        from astrojax.space_weather import load_sw_from_file, get_sw_f107_obs
        sw = load_sw_from_file("path/to/sw19571001.txt")
        val = get_sw_f107_obs(sw, 59569.0)
        ```
    """
    filepath = Path(filepath)
    if not filepath.exists():
        raise FileNotFoundError(f"Space weather file not found: {filepath}")

    (
        mjds,
        kps,
        aps,
        ap_dailys,
        f107_obss,
        f107_adjs,
        f107_obs_ctr81s,
        f107_obs_lst81s,
        f107_adj_ctr81s,
        f107_adj_lst81s,
    ) = parse_cssi_file(str(filepath))

    dtype = get_dtype()

    return SpaceWeatherData(
        mjd=jnp.array(mjds, dtype=dtype),
        kp=jnp.array(kps, dtype=dtype),
        ap=jnp.array(aps, dtype=dtype),
        ap_daily=jnp.array(ap_dailys, dtype=dtype),
        f107_obs=jnp.array(f107_obss, dtype=dtype),
        f107_adj=jnp.array(f107_adjs, dtype=dtype),
        f107_obs_ctr81=jnp.array(f107_obs_ctr81s, dtype=dtype),
        f107_obs_lst81=jnp.array(f107_obs_lst81s, dtype=dtype),
        f107_adj_ctr81=jnp.array(f107_adj_ctr81s, dtype=dtype),
        f107_adj_lst81=jnp.array(f107_adj_lst81s, dtype=dtype),
        mjd_min=jnp.array(mjds[0], dtype=dtype),
        mjd_max=jnp.array(mjds[-1], dtype=dtype),
    )


def load_default_sw() -> SpaceWeatherData:
    """Load the bundled default space weather data (``sw19571001.txt``).

    Uses ``importlib.resources`` to locate the data file bundled with the
    package.

    Returns:
        SpaceWeatherData loaded from the bundled ``sw19571001.txt``.

    Examples:
        ```python
        from astrojax.space_weather import load_default_sw, get_sw_f107_obs
        sw = load_default_sw()
        val = get_sw_f107_obs(sw, 59569.0)
        ```
    """
    data_pkg = importlib.resources.files("astrojax.data.space_weather")
    resource = data_pkg.joinpath("sw19571001.txt")
    with importlib.resources.as_file(resource) as path:
        return load_sw_from_file(path)


def load_cached_sw(
    filepath: str | Path | None = None,
    *,
    max_age_days: float = _DEFAULT_MAX_AGE_DAYS,
) -> SpaceWeatherData:
    """Load space weather data from a local cache, downloading when stale.

    Checks whether the cached file at *filepath* exists and is younger than
    *max_age_days*. If the file is missing or stale, a fresh copy is
    downloaded from CelesTrak. If the download fails or the file cannot be
    parsed, the bundled default data is returned so this function never
    raises on network issues.

    Args:
        filepath: Path to the cached SW file. When ``None`` (the default),
            uses ``<cache_dir>/space_weather/sw19571001.txt``.
        max_age_days: Maximum acceptable age of the cached file in days.
            Defaults to 7.

    Returns:
        SpaceWeatherData loaded from the cached (or freshly downloaded) file,
        or the bundled default data as a fallback.

    Examples:
        ```python
        from astrojax.space_weather import load_cached_sw, get_sw_f107_obs
        sw = load_cached_sw()
        val = get_sw_f107_obs(sw, 59569.0)
        ```
    """
    if filepath is None:
        filepath = get_sw_cache_dir() / _SW_FILENAME
    else:
        filepath = Path(filepath)

    max_age_seconds = max_age_days * 86400.0

    if is_file_stale(filepath, max_age_seconds):
        try:
            download_sw_file(filepath)
        except Exception:
            logger.warning(
                "Failed to download space weather data; falling back to bundled data.",
                exc_info=True,
            )
            return load_default_sw()

    try:
        return load_sw_from_file(filepath)
    except Exception:
        logger.warning(
            "Failed to parse cached space weather file %s; falling back to bundled data.",
            filepath,
            exc_info=True,
        )
        return load_default_sw()



"""
Earth gravity constants for the SGP4/SDP4 propagator.

Provides three standard gravity models: WGS72OLD, WGS72 (standard), and WGS84.
Values match the reference ``sgp4`` Python library exactly.
"""

from math import sqrt
from typing import NamedTuple


class EarthGravity(NamedTuple):
    """Earth gravity model constants for SGP4 propagation.

    Attributes:
        tumin: Time units per minute (1/xke).
        mu: Gravitational parameter [km^3/s^2].
        radiusearthkm: Earth equatorial radius [km].
        xke: Reciprocal of tumin (sqrt(GM) in SGP4 time units).
        j2: Second zonal harmonic.
        j3: Third zonal harmonic.
        j4: Fourth zonal harmonic.
        j3oj2: Ratio j3/j2.
    """

    tumin: float
    mu: float
    radiusearthkm: float
    xke: float
    j2: float
    j3: float
    j4: float
    j3oj2: float


# WGS 72 Old gravity constants
_mu_72old = 398600.79964
_re_72old = 6378.135
_xke_72old = 0.0743669161
_j2_72old = 0.001082616
_j3_72old = -0.00000253881
_j4_72old = -0.00000165597

WGS72OLD = EarthGravity(
    tumin=1.0 / _xke_72old,
    mu=_mu_72old,
    radiusearthkm=_re_72old,
    xke=_xke_72old,
    j2=_j2_72old,
    j3=_j3_72old,
    j4=_j4_72old,
    j3oj2=_j3_72old / _j2_72old,
)
"""WGS 72 Old gravity model (legacy)."""

# WGS 72 gravity constants (standard)
_mu_72 = 398600.8
_re_72 = 6378.135
_xke_72 = 60.0 / sqrt(_re_72**3 / _mu_72)
_j2_72 = 0.001082616
_j3_72 = -0.00000253881
_j4_72 = -0.00000165597

WGS72 = EarthGravity(
    tumin=1.0 / _xke_72,
    mu=_mu_72,
    radiusearthkm=_re_72,
    xke=_xke_72,
    j2=_j2_72,
    j3=_j3_72,
    j4=_j4_72,
    j3oj2=_j3_72 / _j2_72,
)
"""WGS 72 gravity model (standard for SGP4)."""

# WGS 84 gravity constants
_mu_84 = 398600.5
_re_84 = 6378.137
_xke_84 = 60.0 / sqrt(_re_84**3 / _mu_84)
_j2_84 = 0.00108262998905
_j3_84 = -0.00000253215306
_j4_84 = -0.00000161098761

WGS84 = EarthGravity(
    tumin=1.0 / _xke_84,
    mu=_mu_84,
    radiusearthkm=_re_84,
    xke=_xke_84,
    j2=_j2_84,
    j3=_j3_84,
    j4=_j4_84,
    j3oj2=_j3_84 / _j2_84,
)
"""WGS 84 gravity model."""

# Gravity constant lookup by name
GRAVITY_MODELS = {
    "wgs72old": WGS72OLD,
    "wgs72": WGS72,
    "wgs84": WGS84,
}
"""Mapping of gravity model names to ``EarthGravity`` instances."""



"""JAX translations of IAU SOFA routines for Earth orientation modeling.

Implements the IAU 2006/2000A CIO-based precession-nutation model.
Uses routines and computations derived from software provided by SOFA
under license to the user. Does not itself constitute software provided
by and/or endorsed by SOFA.

All functions respect :func:`~astrojax.config.get_dtype` for float precision,
following the same pattern as all other astrojax modules.
"""

from __future__ import annotations

import jax.numpy as jnp
from jax import Array

from astrojax._sofa_nutation_data import LUNI_SOLAR_COEFFS, PLANETARY_COEFFS
from astrojax.attitude_representations import Rx, Ry, Rz
from astrojax.config import get_dtype

# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------

DJ00: float = 2451545.0
"""Julian Date of J2000.0."""

DJC: float = 36525.0
"""Days per Julian century."""

DAS2R: float = 4.848136811095359935899141e-6
"""Arcseconds to radians."""

D2PI: float = 6.283185307179586476925287
"""2*pi."""

TURNAS: float = 1296000.0
"""Arcseconds in a full circle."""

MJD_ZERO: float = 2400000.5
"""Julian Date of MJD zero-point."""

# Units of 0.1 microarcsecond to radians
_U2R: float = DAS2R / 1e7


# ---------------------------------------------------------------------------
# Fundamental arguments (IERS Conventions 2003)
# ---------------------------------------------------------------------------


def fal03(t: Array) -> Array:
    """Mean anomaly of the Moon (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        l in radians.
    """
    return (
        jnp.fmod(
            485868.249036
            + t * (1717915923.2178 + t * (31.8792 + t * (0.051635 + t * (-0.00024470)))),
            TURNAS,
        )
        * DAS2R
    )


def falp03(t: Array) -> Array:
    """Mean anomaly of the Sun (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        l' in radians.
    """
    return (
        jnp.fmod(
            1287104.793048
            + t * (129596581.0481 + t * (-0.5532 + t * (0.000136 + t * (-0.00001149)))),
            TURNAS,
        )
        * DAS2R
    )


def faf03(t: Array) -> Array:
    """Mean argument of the latitude of the Moon (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        F in radians.
    """
    return (
        jnp.fmod(
            335779.526232
            + t * (1739527262.8478 + t * (-12.7512 + t * (-0.001037 + t * (0.00000417)))),
            TURNAS,
        )
        * DAS2R
    )


def fad03(t: Array) -> Array:
    """Mean elongation of the Moon from the Sun (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        D in radians.
    """
    return (
        jnp.fmod(
            1072260.703692
            + t * (1602961601.2090 + t * (-6.3706 + t * (0.006593 + t * (-0.00003169)))),
            TURNAS,
        )
        * DAS2R
    )


def faom03(t: Array) -> Array:
    """Mean longitude of the Moon's ascending node (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Omega in radians.
    """
    return (
        jnp.fmod(
            450160.398036 + t * (-6962890.5431 + t * (7.4722 + t * (0.007702 + t * (-0.00005939)))),
            TURNAS,
        )
        * DAS2R
    )


def fame03(t: Array) -> Array:
    """Mean longitude of Mercury (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Mean longitude in radians.
    """
    return jnp.fmod(4.402608842 + 2608.7903141574 * t, D2PI)


def fave03(t: Array) -> Array:
    """Mean longitude of Venus (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Mean longitude in radians.
    """
    return jnp.fmod(3.176146697 + 1021.3285546211 * t, D2PI)


def fae03(t: Array) -> Array:
    """Mean longitude of Earth (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Mean longitude in radians.
    """
    return jnp.fmod(1.753470314 + 628.3075849991 * t, D2PI)


def fama03(t: Array) -> Array:
    """Mean longitude of Mars (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Mean longitude in radians.
    """
    return jnp.fmod(6.203480913 + 334.0612426700 * t, D2PI)


def faju03(t: Array) -> Array:
    """Mean longitude of Jupiter (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Mean longitude in radians.
    """
    return jnp.fmod(0.599546497 + 52.9690962641 * t, D2PI)


def fasa03(t: Array) -> Array:
    """Mean longitude of Saturn (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Mean longitude in radians.
    """
    return jnp.fmod(0.874016757 + 21.3299104960 * t, D2PI)


def faur03(t: Array) -> Array:
    """Mean longitude of Uranus (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Mean longitude in radians.
    """
    return jnp.fmod(5.481293872 + 7.4781598567 * t, D2PI)


def fane03(t: Array) -> Array:
    """Mean longitude of Neptune (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        Mean longitude in radians.
    """
    return jnp.fmod(5.311886287 + 3.8133035638 * t, D2PI)


def fapa03(t: Array) -> Array:
    """General accumulated precession in longitude (IERS 2003).

    Args:
        t: TDB Julian centuries since J2000.0.

    Returns:
        General precession in radians.
    """
    return (0.024381750 + 0.00000538691 * t) * t


# ---------------------------------------------------------------------------
# Mean obliquity
# ---------------------------------------------------------------------------


def obl06(date1: Array, date2: Array) -> Array:
    """Mean obliquity of the ecliptic, IAU 2006 precession.

    Args:
        date1: TT as 2-part Julian Date (part 1).
        date2: TT as 2-part Julian Date (part 2).

    Returns:
        Obliquity of the ecliptic in radians.
    """
    t = ((date1 - DJ00) + date2) / DJC
    eps0 = 84381.406 + t * (
        -46.836769 + t * (-0.0001831 + t * (0.00200340 + t * (-0.000000576 + t * (-0.0000000434))))
    )
    return eps0 * DAS2R


# ---------------------------------------------------------------------------
# Fukushima-Williams angles
# ---------------------------------------------------------------------------


def pfw06(date1: Array, date2: Array) -> tuple[Array, Array, Array, Array]:
    """Precession angles, IAU 2006, Fukushima-Williams 4-angle formulation.

    Args:
        date1: TT as 2-part Julian Date (part 1).
        date2: TT as 2-part Julian Date (part 2).

    Returns:
        Tuple of (gamb, phib, psib, epsa) in radians.
    """
    t = ((date1 - DJ00) + date2) / DJC

    gamb = (
        -0.052928
        + t
        * (
            10.556378
            + t * (0.4932044 + t * (-0.00031238 + t * (-0.000002788 + t * (0.0000000260))))
        )
    ) * DAS2R

    phib = (
        84381.412819
        + t
        * (
            -46.811016
            + t * (0.0511268 + t * (0.00053289 + t * (-0.000000440 + t * (-0.0000000176))))
        )
    ) * DAS2R

    psib = (
        -0.041775
        + t
        * (
            5038.481484
            + t * (1.5584175 + t * (-0.00018522 + t * (-0.000026452 + t * (-0.0000000148))))
        )
    ) * DAS2R

    epsa = obl06(date1, date2)

    return gamb, phib, psib, epsa


# ---------------------------------------------------------------------------
# Nutation IAU 2000A
# ---------------------------------------------------------------------------


def nut00a(date1: Array, date2: Array) -> tuple[Array, Array]:
    """Nutation, IAU 2000A model (MHB2000 luni-solar and planetary).

    Vectorized implementation using matmul over all 678 luni-solar and
    687 planetary nutation terms.

    Args:
        date1: TT as 2-part Julian Date (part 1).
        date2: TT as 2-part Julian Date (part 2).

    Returns:
        Tuple of (dpsi, deps) nutation in longitude and obliquity [radians].
    """
    dtype = get_dtype()
    t = dtype(((date1 - DJ00) + date2) / DJC)

    # Fundamental (Delaunay) arguments
    el = fal03(t)
    elp = falp03(t)
    f = faf03(t)
    d = fad03(t)
    om = faom03(t)
    delaunay = jnp.array([el, elp, f, d, om])

    # Planetary longitudes
    al = fal03(t)
    af = faf03(t)
    ad = fad03(t)
    aom = faom03(t)
    ame = fame03(t)
    ave = fave03(t)
    ae = fae03(t)
    ama = fama03(t)
    aju = faju03(t)
    asa = fasa03(t)
    aur = faur03(t)
    ane = fane03(t)
    apa = fapa03(t)
    planetary = jnp.array([al, af, ad, aom, ame, ave, ae, ama, aju, asa, aur, ane, apa])

    # ---- Luni-solar nutation (678 terms) ----
    ls = jnp.array(LUNI_SOLAR_COEFFS, dtype=dtype)
    ls_nfa = ls[:, :5]  # (678, 5) integer multipliers
    ls_sp = ls[:, 5]  # sine coefficient for longitude
    ls_spt = ls[:, 6]  # time-dependent sine coefficient
    ls_cp = ls[:, 7]  # cosine coefficient for longitude
    ls_ce = ls[:, 8]  # cosine coefficient for obliquity
    ls_cet = ls[:, 9]  # time-dependent cosine coefficient
    ls_se = ls[:, 10]  # sine coefficient for obliquity

    # All arguments at once: (678, 5) @ (5,) -> (678,)
    ls_args = ls_nfa @ delaunay
    ls_sin = jnp.sin(ls_args)
    ls_cos = jnp.cos(ls_args)

    dpsi_ls = jnp.sum((ls_sp + ls_spt * t) * ls_sin + ls_cp * ls_cos)
    deps_ls = jnp.sum((ls_ce + ls_cet * t) * ls_cos + ls_se * ls_sin)

    # ---- Planetary nutation (687 terms) ----
    pl = jnp.array(PLANETARY_COEFFS, dtype=dtype)
    pl_nfa = pl[:, :13]  # (687, 13) integer multipliers
    pl_sp = pl[:, 13]  # sine coefficient for longitude
    pl_cp = pl[:, 14]  # cosine coefficient for longitude
    pl_se = pl[:, 15]  # sine coefficient for obliquity
    pl_ce = pl[:, 16]  # cosine coefficient for obliquity

    # All arguments at once: (687, 13) @ (13,) -> (687,)
    pl_args = pl_nfa @ planetary
    pl_sin = jnp.sin(pl_args)
    pl_cos = jnp.cos(pl_args)

    dpsi_pl = jnp.sum(pl_sp * pl_sin + pl_cp * pl_cos)
    deps_pl = jnp.sum(pl_se * pl_sin + pl_ce * pl_cos)

    # Total nutation (convert from 0.1 uas to radians)
    dpsi = (dpsi_ls + dpsi_pl) * _U2R
    deps = (deps_ls + deps_pl) * _U2R

    return dpsi, deps


# ---------------------------------------------------------------------------
# Nutation IAU 2006/2000A (P03 corrected)
# ---------------------------------------------------------------------------


def nut06a(date1: Array, date2: Array) -> tuple[Array, Array]:
    """Nutation, IAU 2006/2000A (with P03 precession adjustment).

    Applies the P03-compatible J2 correction factor to the IAU 2000A nutation.

    Args:
        date1: TT as 2-part Julian Date (part 1).
        date2: TT as 2-part Julian Date (part 2).

    Returns:
        Tuple of (dpsi, deps) nutation in longitude and obliquity [radians].
    """
    t = ((date1 - DJ00) + date2) / DJC

    # J2 correction factor for P03 precession
    fj2 = -2.7774e-6 * t

    # IAU 2000A nutation
    dp, de = nut00a(date1, date2)

    # Apply P03 adjustments
    dpsi = dp * (1.0 + 0.4697e-6 * fj2)
    deps = de * (1.0 + fj2)

    return dpsi, deps


# ---------------------------------------------------------------------------
# Fukushima-Williams angles to rotation matrix
# ---------------------------------------------------------------------------


def fw2m(gamb: Array, phib: Array, psi: Array, eps: Array) -> Array:
    """Fukushima-Williams angles to rotation matrix.

    Form the rotation matrix given the Fukushima-Williams angles:
    ``NxPxB = R_1(-eps) . R_3(-psi) . R_1(phib) . R_3(gamb)``

    Where R_1 and R_3 use the same convention as Rx/Rz
    (SOFA/ERFA convention with +sin in off-diagonal).

    Args:
        gamb: F-W angle gamma_bar (radians).
        phib: F-W angle phi_bar (radians).
        psi: F-W angle psi (radians).
        eps: F-W angle epsilon (radians).

    Returns:
        3x3 rotation matrix (NPB matrix).
    """
    return Rx(-eps) @ Rz(-psi) @ Rx(phib) @ Rz(gamb)


# ---------------------------------------------------------------------------
# Precession-nutation matrix
# ---------------------------------------------------------------------------


def pnm06a(date1: Array, date2: Array) -> Array:
    """Form the bias-precession-nutation matrix, IAU 2006/2000A.

    Combines Fukushima-Williams precession angles with IAU 2006/2000A
    nutation to form the complete GCRS-to-true (BPN) rotation matrix.

    Args:
        date1: TT as 2-part Julian Date (part 1).
        date2: TT as 2-part Julian Date (part 2).

    Returns:
        3x3 bias-precession-nutation matrix.
    """
    # Fukushima-Williams angles for frame bias and precession
    gamb, phib, psib, epsa = pfw06(date1, date2)

    # Nutation components
    dpsi, deps = nut06a(date1, date2)

    # Form the matrix: FW angles with nutation added
    return fw2m(gamb, phib, psib + dpsi, epsa + deps)


# ---------------------------------------------------------------------------
# Extract CIP coordinates from BPN matrix
# ---------------------------------------------------------------------------


def bpn2xy(rbpn: Array) -> tuple[Array, Array]:
    """Extract CIP X, Y coordinates from the BPN matrix.

    Args:
        rbpn: 3x3 bias-precession-nutation matrix.

    Returns:
        Tuple of (x, y) CIP coordinates.
    """
    return rbpn[2, 0], rbpn[2, 1]


# ---------------------------------------------------------------------------
# CIO locator s
# ---------------------------------------------------------------------------

# Polynomial coefficients for s + XY/2 (arcseconds -> radians at evaluation)
_S06_SP = (94.00e-6, 3808.65e-6, -122.68e-6, -72574.11e-6, 27.98e-6, 15.62e-6)

# S06 CIO locator series coefficient data.
# Stored as Python tuples to avoid module-level jnp.array calls
# (which would require x64 to be enabled at import time).
# Each group: (nfa_tuple, sc_tuple) where nfa has 8-element integer rows
# and sc has 2-element (sin, cos) coefficient rows.

# fmt: off
# Terms of order t^0 (33 terms)
_S06_S0_NFA_DATA = (
    (0,0,0,0,1,0,0,0), (0,0,0,0,2,0,0,0), (0,0,2,-2,3,0,0,0),
    (0,0,2,-2,1,0,0,0), (0,0,2,-2,2,0,0,0), (0,0,2,0,3,0,0,0),
    (0,0,2,0,1,0,0,0), (0,0,0,0,3,0,0,0), (0,1,0,0,1,0,0,0),
    (0,1,0,0,-1,0,0,0), (1,0,0,0,-1,0,0,0), (1,0,0,0,1,0,0,0),
    (0,1,2,-2,3,0,0,0), (0,1,2,-2,1,0,0,0), (0,0,4,-4,4,0,0,0),
    (0,0,1,-1,1,-8,12,0), (0,0,2,0,0,0,0,0), (0,0,2,0,2,0,0,0),
    (1,0,2,0,3,0,0,0), (1,0,2,0,1,0,0,0), (0,0,2,-2,0,0,0,0),
    (0,1,-2,2,-3,0,0,0), (0,1,-2,2,-1,0,0,0), (0,0,0,0,0,8,-13,-1),
    (0,0,0,2,0,0,0,0), (2,0,-2,0,-1,0,0,0), (0,1,2,-2,2,0,0,0),
    (1,0,0,-2,1,0,0,0), (1,0,0,-2,-1,0,0,0), (0,0,4,-2,4,0,0,0),
    (0,0,2,-2,4,0,0,0), (1,0,-2,0,-3,0,0,0), (1,0,-2,0,-1,0,0,0),
)
_S06_S0_SC_DATA = (
    (-2640.73e-6, 0.39e-6), (-63.53e-6, 0.02e-6), (-11.75e-6, -0.01e-6),
    (-11.21e-6, -0.01e-6), (4.57e-6, 0.00e-6), (-2.02e-6, 0.00e-6),
    (-1.98e-6, 0.00e-6), (1.72e-6, 0.00e-6), (1.41e-6, 0.01e-6),
    (1.26e-6, 0.01e-6), (0.63e-6, 0.00e-6), (0.63e-6, 0.00e-6),
    (-0.46e-6, 0.00e-6), (-0.45e-6, 0.00e-6), (-0.36e-6, 0.00e-6),
    (0.24e-6, 0.12e-6), (-0.32e-6, 0.00e-6), (-0.28e-6, 0.00e-6),
    (-0.27e-6, 0.00e-6), (-0.26e-6, 0.00e-6), (0.21e-6, 0.00e-6),
    (-0.19e-6, 0.00e-6), (-0.18e-6, 0.00e-6), (0.10e-6, -0.05e-6),
    (-0.15e-6, 0.00e-6), (0.14e-6, 0.00e-6), (0.14e-6, 0.00e-6),
    (-0.14e-6, 0.00e-6), (-0.14e-6, 0.00e-6), (-0.13e-6, 0.00e-6),
    (0.11e-6, 0.00e-6), (-0.11e-6, 0.00e-6), (-0.11e-6, 0.00e-6),
)

# Terms of order t^1 (3 terms)
_S06_S1_NFA_DATA = (
    (0,0,0,0,2,0,0,0), (0,0,0,0,1,0,0,0), (0,0,2,-2,3,0,0,0),
)
_S06_S1_SC_DATA = (
    (-0.07e-6, 3.57e-6), (1.73e-6, -0.03e-6), (0.00e-6, 0.48e-6),
)

# Terms of order t^2 (25 terms)
_S06_S2_NFA_DATA = (
    (0,0,0,0,1,0,0,0), (0,0,2,-2,2,0,0,0), (0,0,2,0,2,0,0,0),
    (0,0,0,0,2,0,0,0), (0,1,0,0,0,0,0,0), (1,0,0,0,0,0,0,0),
    (0,1,2,-2,2,0,0,0), (0,0,2,0,1,0,0,0), (1,0,2,0,2,0,0,0),
    (0,1,-2,2,-2,0,0,0), (1,0,0,-2,0,0,0,0), (0,0,2,-2,1,0,0,0),
    (1,0,-2,0,-2,0,0,0), (0,0,0,2,0,0,0,0), (1,0,0,0,1,0,0,0),
    (1,0,-2,-2,-2,0,0,0), (1,0,0,0,-1,0,0,0), (1,0,2,0,1,0,0,0),
    (2,0,0,-2,0,0,0,0), (2,0,-2,0,-1,0,0,0), (0,0,2,2,2,0,0,0),
    (2,0,2,0,2,0,0,0), (2,0,0,0,0,0,0,0), (1,0,2,-2,2,0,0,0),
    (0,0,2,0,0,0,0,0),
)
_S06_S2_SC_DATA = (
    (743.52e-6, -0.17e-6), (56.91e-6, 0.06e-6), (9.84e-6, -0.01e-6),
    (-8.85e-6, 0.01e-6), (-6.38e-6, -0.05e-6), (-3.07e-6, 0.00e-6),
    (2.23e-6, 0.00e-6), (1.67e-6, 0.00e-6), (1.30e-6, 0.00e-6),
    (0.93e-6, 0.00e-6), (0.68e-6, 0.00e-6), (-0.55e-6, 0.00e-6),
    (0.53e-6, 0.00e-6), (-0.27e-6, 0.00e-6), (-0.27e-6, 0.00e-6),
    (-0.26e-6, 0.00e-6), (-0.25e-6, 0.00e-6), (0.22e-6, 0.00e-6),
    (-0.21e-6, 0.00e-6), (0.20e-6, 0.00e-6), (0.17e-6, 0.00e-6),
    (0.13e-6, 0.00e-6), (-0.13e-6, 0.00e-6), (-0.12e-6, 0.00e-6),
    (-0.11e-6, 0.00e-6),
)

# Terms of order t^3 (4 terms)
_S06_S3_NFA_DATA = (
    (0,0,0,0,1,0,0,0), (0,0,2,-2,2,0,0,0),
    (0,0,2,0,2,0,0,0), (0,0,0,0,2,0,0,0),
)
_S06_S3_SC_DATA = (
    (0.30e-6, -23.42e-6), (-0.03e-6, -1.46e-6),
    (-0.01e-6, -0.25e-6), (0.00e-6, 0.23e-6),
)

# Terms of order t^4 (1 term)
_S06_S4_NFA_DATA = ((0,0,0,0,1,0,0,0),)
_S06_S4_SC_DATA = ((-0.26e-6, -0.01e-6),)
# fmt: on


def _s06_series(nfa: Array, sc: Array, fa: Array) -> Array:
    """Evaluate one order of the CIO locator series.

    Args:
        nfa: Integer multiplier array, shape (N, 8).
        sc: Sine/cosine coefficient array, shape (N, 2).
        fa: Fundamental arguments, shape (8,).

    Returns:
        Sum of s*sin(a) + c*cos(a) over all terms.
    """
    args = nfa @ fa  # (N,)
    return jnp.sum(sc[:, 0] * jnp.sin(args) + sc[:, 1] * jnp.cos(args))


def s06(date1: Array, date2: Array, x: Array, y: Array) -> Array:
    """CIO locator s, positioning the Celestial Intermediate Origin on the
    equator of the CIP. Compatible with IAU 2006/2000A precession-nutation.

    The series is actually for s + XY/2. The function subtracts XY/2 to
    return s itself.

    Args:
        date1: TT as 2-part Julian Date (part 1).
        date2: TT as 2-part Julian Date (part 2).
        x: CIP x coordinate.
        y: CIP y coordinate.

    Returns:
        CIO locator s in radians.
    """
    dtype = get_dtype()
    t = dtype(((date1 - DJ00) + date2) / DJC)

    # Fundamental arguments
    fa = jnp.array(
        [
            fal03(t),  # l
            falp03(t),  # l'
            faf03(t),  # F
            fad03(t),  # D
            faom03(t),  # Om
            fave03(t),  # LVe
            fae03(t),  # LE
            fapa03(t),  # pA
        ]
    )

    # Create coefficient arrays from tuple data at runtime
    s0_nfa = jnp.array(_S06_S0_NFA_DATA, dtype=dtype)
    s0_sc = jnp.array(_S06_S0_SC_DATA, dtype=dtype)
    s1_nfa = jnp.array(_S06_S1_NFA_DATA, dtype=dtype)
    s1_sc = jnp.array(_S06_S1_SC_DATA, dtype=dtype)
    s2_nfa = jnp.array(_S06_S2_NFA_DATA, dtype=dtype)
    s2_sc = jnp.array(_S06_S2_SC_DATA, dtype=dtype)
    s3_nfa = jnp.array(_S06_S3_NFA_DATA, dtype=dtype)
    s3_sc = jnp.array(_S06_S3_SC_DATA, dtype=dtype)
    s4_nfa = jnp.array(_S06_S4_NFA_DATA, dtype=dtype)
    s4_sc = jnp.array(_S06_S4_SC_DATA, dtype=dtype)

    # Evaluate series at each power of t
    w0 = dtype(_S06_SP[0]) + _s06_series(s0_nfa, s0_sc, fa)
    w1 = dtype(_S06_SP[1]) + _s06_series(s1_nfa, s1_sc, fa)
    w2 = dtype(_S06_SP[2]) + _s06_series(s2_nfa, s2_sc, fa)
    w3 = dtype(_S06_SP[3]) + _s06_series(s3_nfa, s3_sc, fa)
    w4 = dtype(_S06_SP[4]) + _s06_series(s4_nfa, s4_sc, fa)
    w5 = dtype(_S06_SP[5])

    # Horner form + convert arcseconds to radians, subtract XY/2
    s = (w0 + (w1 + (w2 + (w3 + (w4 + w5 * t) * t) * t) * t) * t) * DAS2R - x * y / 2.0

    return s


# ---------------------------------------------------------------------------
# Composite: X, Y, s
# ---------------------------------------------------------------------------


def xys06a(date1: Array, date2: Array) -> tuple[Array, Array, Array]:
    """CIP X, Y coordinates and CIO locator s, IAU 2006/2000A.

    Combines pnm06a, bpn2xy, and s06 to return the full set of
    CIO-based parameters.

    Args:
        date1: TT as 2-part Julian Date (part 1).
        date2: TT as 2-part Julian Date (part 2).

    Returns:
        Tuple of (x, y, s) where x,y are CIP coordinates and s is the
        CIO locator, all in radians.
    """
    # Form the bias-precession-nutation matrix
    rbpn = pnm06a(date1, date2)

    # Extract CIP X, Y
    x, y = bpn2xy(rbpn)

    # CIO locator s
    s_val = s06(date1, date2, x, y)

    return x, y, s_val


# ---------------------------------------------------------------------------
# CIP to celestial-to-intermediate matrix
# ---------------------------------------------------------------------------


def c2ixys(x: Array, y: Array, s: Array) -> Array:
    """Form the celestial-to-intermediate matrix given CIP X, Y and CIO locator s.

    Uses the method: ``Rz(-(e+s)) @ Ry(d) @ Rz(e)`` where
    ``d = arctan(sqrt((x^2 + y^2) / (1 - x^2 - y^2)))``
    and ``e = atan2(y, x)``.

    This matches the SOFA iauC2ixys implementation exactly (astrojax and SOFA
    use identical Rx/Ry/Rz conventions).

    Args:
        x: CIP x coordinate.
        y: CIP y coordinate.
        s: CIO locator.

    Returns:
        3x3 celestial-to-intermediate matrix.
    """
    # Derived quantities
    r2 = x * x + y * y
    e = jnp.where(r2 > 0.0, jnp.arctan2(y, x), 0.0)
    d = jnp.arctan(jnp.sqrt(r2 / (1.0 - r2)))

    return Rz(-(e + s)) @ Ry(d) @ Rz(e)


# ---------------------------------------------------------------------------
# Earth Rotation Angle
# ---------------------------------------------------------------------------


def era00(dj1: Array, dj2: Array) -> Array:
    """Earth Rotation Angle (IAU 2000 model).

    Args:
        dj1: UT1 as 2-part Julian Date (part 1).
        dj2: UT1 as 2-part Julian Date (part 2).

    Returns:
        Earth Rotation Angle in radians (0 to 2*pi).
    """
    # Days since J2000.0
    t = dj1 + dj2 - DJ00

    # Fractional part of dj1 + dj2
    f = jnp.fmod(dj1, 1.0) + jnp.fmod(dj2, 1.0)

    # Earth Rotation Angle
    theta = (
        jnp.fmod(
            f + 0.7790572732640 + 0.00273781191135448 * t,
            1.0,
        )
        * D2PI
    )

    return theta


# ---------------------------------------------------------------------------
# TIO locator s'
# ---------------------------------------------------------------------------


def sp00(date1: Array, date2: Array) -> Array:
    """TIO locator s', positioning the Terrestrial Intermediate Origin.

    The dominant terms of s' can be approximated as ``-47e-6 * t * DAS2R``
    where t is Julian centuries of TT since J2000.0.

    Args:
        date1: TT as 2-part Julian Date (part 1).
        date2: TT as 2-part Julian Date (part 2).

    Returns:
        TIO locator s' in radians.
    """
    t = ((date1 - DJ00) + date2) / DJC
    return -47e-6 * t * DAS2R


# ---------------------------------------------------------------------------
# Polar motion matrix
# ---------------------------------------------------------------------------


def pom00(xp: Array, yp: Array, sp: Array) -> Array:
    """Form the polar motion matrix (TIRS -> ITRS).

    The matrix is ``Rx(-yp) @ Ry(-xp) @ Rz(sp)``.

    This matches the SOFA iauPom00 implementation exactly (astrojax and SOFA
    use identical Rx/Ry/Rz conventions).

    Args:
        xp: Polar motion x-component (radians, positive towards Greenwich).
        yp: Polar motion y-component (radians, positive towards 270E).
        sp: TIO locator s' (radians).

    Returns:
        3x3 polar motion matrix.
    """
    return Rx(-yp) @ Ry(-xp) @ Rz(sp)



"""Frame transformations.

This sub-module provides functions for converting between common
coordinate frames used in astrodynamics:

- **GCRF-ITRF transformations**: converting between the Geocentric Celestial
  Reference Frame and the International Terrestrial Reference Frame using the
  full IAU 2006/2000A CIO-based model (bias-precession-nutation, Earth rotation
  angle, and polar motion).
- **TEME transformations**: converting SGP4 output (TEME frame) to PEF, ITRF,
  and GCRF via GMST rotation and polar motion.
- **Ecliptic-ICRF transformations**: converting between the ecliptic
  coordinate frame and ICRF via a fixed rotation by the J2000 mean obliquity.
- **ECI-ECEF aliases**: backward-compatible names mapping to GCRF/ITRF functions.
"""

from .ecliptic import (
    rotation_ecliptic_to_icrf,
    rotation_icrf_to_ecliptic,
    state_ecliptic_to_icrf,
    state_icrf_to_ecliptic,
)
from .gcrf_itrf import (
    bias_precession_nutation,
    earth_rotation,
    earth_rotation_angle,
    polar_motion,
    rotation_ecef_to_eci,
    rotation_eci_to_ecef,
    rotation_gcrf_to_itrf,
    rotation_itrf_to_gcrf,
    state_ecef_to_eci,
    state_eci_to_ecef,
    state_gcrf_to_itrf,
    state_itrf_to_gcrf,
)
from .teme import (
    rotation_itrf_to_teme,
    rotation_pef_to_teme,
    rotation_teme_to_itrf,
    rotation_teme_to_pef,
    state_gcrf_to_teme,
    state_itrf_to_teme,
    state_pef_to_teme,
    state_teme_to_gcrf,
    state_teme_to_itrf,
    state_teme_to_pef,
)

__all__ = [
    # GCRF/ITRF (primary API)
    "bias_precession_nutation",
    "earth_rotation_angle",
    "polar_motion",
    "rotation_gcrf_to_itrf",
    "rotation_itrf_to_gcrf",
    "state_gcrf_to_itrf",
    "state_itrf_to_gcrf",
    # TEME transformations
    "rotation_teme_to_pef",
    "rotation_pef_to_teme",
    "rotation_teme_to_itrf",
    "rotation_itrf_to_teme",
    "state_teme_to_pef",
    "state_pef_to_teme",
    "state_teme_to_itrf",
    "state_itrf_to_teme",
    "state_teme_to_gcrf",
    "state_gcrf_to_teme",
    # Ecliptic/ICRF transformations
    "rotation_ecliptic_to_icrf",
    "rotation_icrf_to_ecliptic",
    "state_ecliptic_to_icrf",
    "state_icrf_to_ecliptic",
    # ECI/ECEF aliases
    "earth_rotation",
    "rotation_eci_to_ecef",
    "rotation_ecef_to_eci",
    "state_eci_to_ecef",
    "state_ecef_to_eci",
]



"""SpaceTrack HTTP client with authentication and query execution.

Handles session management via cookie-based authentication against
the Space-Track.org API.
"""

from __future__ import annotations

import json
import time
from typing import Any

import httpx

from astrojax._gp_record import GPRecord
from astrojax.spacetrack._query import SpaceTrackQuery
from astrojax.spacetrack._rate_limiter import RateLimitConfig, RateLimiter
from astrojax.spacetrack._responses import SATCATRecord

_DEFAULT_BASE_URL = "https://www.space-track.org"


class SpaceTrackClient:
    """SpaceTrack API client with session-based authentication.

    Lazily authenticates on first query and re-authenticates on session expiry.

    Args:
        identity: Space-Track.org login email.
        password: Space-Track.org password.
        base_url: Custom base URL for testing.
        rate_limit: Rate limit configuration.
    """

    def __init__(
        self,
        identity: str,
        password: str,
        base_url: str | None = None,
        *,
        rate_limit: RateLimitConfig | None = None,
    ) -> None:
        self._identity = identity
        self._password = password
        self._base_url = (base_url or _DEFAULT_BASE_URL).rstrip("/")
        self._client = httpx.Client(timeout=120.0, follow_redirects=True)
        self._authenticated = False
        config = rate_limit if rate_limit is not None else RateLimitConfig()
        self._rate_limiter = RateLimiter(config)

    def _wait_for_rate_limit(self) -> None:
        """Wait for rate limit clearance before making a request."""
        wait = self._rate_limiter.acquire()
        if wait > 0:
            time.sleep(wait)

    def authenticate(self) -> None:
        """Explicitly authenticate with Space-Track.org.

        Called automatically on first query. Call explicitly to verify
        credentials early.

        Raises:
            RuntimeError: If authentication fails.
        """
        self._wait_for_rate_limit()

        url = f"{self._base_url}/ajaxauth/login"
        response = self._client.post(
            url,
            data={"identity": self._identity, "password": self._password},
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )
        response.raise_for_status()

        body = response.text
        if '"Login":"Failed"' in body or '"Login": "Failed"' in body:
            raise RuntimeError(
                "SpaceTrack authentication failed: invalid credentials"
            )

        self._authenticated = True

    def _ensure_authenticated(self) -> None:
        """Ensure the client is authenticated, authenticating if necessary."""
        if not self._authenticated:
            self.authenticate()

    def _execute_get(self, url: str) -> str:
        """Execute an HTTP GET request and return the response body as text."""
        self._wait_for_rate_limit()
        response = self._client.get(url)
        response.raise_for_status()
        return response.text

    def _authenticated_get_string(self, url: str) -> str:
        """Execute an authenticated GET, re-authenticating on 401."""
        self._ensure_authenticated()
        try:
            return self._execute_get(url)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 401:
                self.authenticate()
                return self._execute_get(url)
            raise

    # ========================================
    # Query methods
    # ========================================

    def query_raw(self, query: SpaceTrackQuery) -> str:
        """Execute a query and return the raw response body as a string.

        Args:
            query: A SpaceTrackQuery instance.

        Returns:
            Raw response body.
        """
        url = f"{self._base_url}{query.build()}"
        return self._authenticated_get_string(url)

    def query_json(self, query: SpaceTrackQuery) -> list[dict[str, Any]]:
        """Execute a query and return parsed JSON values.

        Args:
            query: A SpaceTrackQuery instance.

        Returns:
            List of JSON dictionaries.

        Raises:
            RuntimeError: If the query format is not JSON.
        """
        if not query.output_format().is_json():
            raise RuntimeError("query_json requires JSON output format")
        body = self.query_raw(query)
        return json.loads(body)

    def query_gp(self, query: SpaceTrackQuery) -> list[GPRecord]:
        """Execute a GP query and return typed GP records.

        Args:
            query: A SpaceTrackQuery instance.

        Returns:
            List of GP records.

        Raises:
            RuntimeError: If the query format is not JSON.
        """
        if not query.output_format().is_json():
            raise RuntimeError("query_gp requires JSON output format")
        body = self.query_raw(query)
        records_json = json.loads(body)
        return [GPRecord.from_json_dict(r) for r in records_json]

    def query_satcat(self, query: SpaceTrackQuery) -> list[SATCATRecord]:
        """Execute a SATCAT query and return typed SATCAT records.

        Args:
            query: A SpaceTrackQuery instance.

        Returns:
            List of SATCAT records.

        Raises:
            RuntimeError: If the query format is not JSON.
        """
        if not query.output_format().is_json():
            raise RuntimeError("query_satcat requires JSON output format")
        body = self.query_raw(query)
        records_json = json.loads(body)
        return [SATCATRecord.from_json_dict(r) for r in records_json]



"""Orbit dynamics force models for perturbation modelling.

Provides gravitational, atmospheric, and radiation force models for
orbit propagation:

- **Ephemerides**: Low-precision Sun and Moon position (Montenbruck & Gill)
- **Planetary ephemerides**: Approximate heliocentric positions (JPL Table 1)
- **Gravity**: Point-mass gravitational acceleration
- **Third body**: Sun and Moon gravitational perturbations
- **Density**: Harris-Priester and NRLMSISE-00 atmospheric density models
- **Drag**: Atmospheric drag acceleration
- **SRP**: Solar radiation pressure and eclipse shadow models
"""

from ._jpl_planetary_coefficients import (
    EMB_ID,
    JUPITER_ID,
    MARS_ID,
    MERCURY_ID,
    NEPTUNE_ID,
    SATURN_ID,
    URANUS_ID,
    VENUS_ID,
)
from .config import ForceModelConfig, SpacecraftParams
from .density import density_harris_priester
from .drag import accel_drag
from .ephemerides import moon_position, sun_position
from .factory import create_orbit_dynamics
from .gravity import (
    GravityModel,
    accel_gravity,
    accel_gravity_spherical_harmonics,
    accel_point_mass,
)
from .nrlmsise00 import density_nrlmsise00, density_nrlmsise00_geod
from .planetary_ephemerides import (
    emb_position_jpl_approx,
    jupiter_position_jpl_approx,
    mars_position_jpl_approx,
    mercury_position_jpl_approx,
    neptune_position_jpl_approx,
    planet_position_jpl_approx,
    saturn_position_jpl_approx,
    uranus_position_jpl_approx,
    venus_position_jpl_approx,
)
from .srp import accel_srp, eclipse_conical, eclipse_cylindrical
from .third_body import accel_third_body_moon, accel_third_body_sun

__all__ = [
    # Ephemerides
    "sun_position",
    "moon_position",
    # Planetary Ephemerides - Planet IDs
    "MERCURY_ID",
    "VENUS_ID",
    "EMB_ID",
    "MARS_ID",
    "JUPITER_ID",
    "SATURN_ID",
    "URANUS_ID",
    "NEPTUNE_ID",
    # Planetary Ephemerides - Position functions
    "mercury_position_jpl_approx",
    "venus_position_jpl_approx",
    "emb_position_jpl_approx",
    "mars_position_jpl_approx",
    "jupiter_position_jpl_approx",
    "saturn_position_jpl_approx",
    "uranus_position_jpl_approx",
    "neptune_position_jpl_approx",
    "planet_position_jpl_approx",
    # Gravity
    "accel_point_mass",
    "accel_gravity",
    "GravityModel",
    "accel_gravity_spherical_harmonics",
    # Third body
    "accel_third_body_sun",
    "accel_third_body_moon",
    # Density
    "density_harris_priester",
    "density_nrlmsise00",
    "density_nrlmsise00_geod",
    # Drag
    "accel_drag",
    # SRP & Eclipse
    "accel_srp",
    "eclipse_conical",
    "eclipse_cylindrical",
    # Dynamics factory
    "ForceModelConfig",
    "SpacecraftParams",
    "create_orbit_dynamics",
]



"""Sliding-window rate limiter for SpaceTrack API requests.

Space-Track.org enforces rate limits of 30 requests per minute and
300 requests per hour. Default limits are set conservatively at ~83%
(25/min, 250/hour).
"""

from __future__ import annotations

import threading
import time
from collections import deque
from dataclasses import dataclass

_U32_MAX = 2**32 - 1


@dataclass
class RateLimitConfig:
    """Configuration for SpaceTrack API rate limiting.

    Args:
        max_per_minute: Maximum requests per rolling 60-second window.
        max_per_hour: Maximum requests per rolling 3600-second window.
    """

    max_per_minute: int = 25
    max_per_hour: int = 250

    @classmethod
    def disabled(cls) -> RateLimitConfig:
        """Create a configuration that disables rate limiting.

        Returns:
            A RateLimitConfig with effectively unlimited limits.
        """
        return cls(max_per_minute=_U32_MAX, max_per_hour=_U32_MAX)

    def __str__(self) -> str:
        return (
            f"RateLimitConfig(max_per_minute={self.max_per_minute}, "
            f"max_per_hour={self.max_per_hour})"
        )

    def __repr__(self) -> str:
        return self.__str__()


class RateLimiter:
    """Sliding-window rate limiter that tracks request timestamps.

    Thread-safe via internal lock.

    Args:
        config: Rate limit configuration.
    """

    def __init__(self, config: RateLimitConfig) -> None:
        self._config = config
        self._minute_window: deque[float] = deque()
        self._hour_window: deque[float] = deque()
        self._lock = threading.Lock()

    def acquire(self) -> float:
        """Acquire permission to make a request.

        Returns the number of seconds the caller must sleep before
        proceeding. Zero means proceed immediately.

        Returns:
            Wait time in seconds (0.0 means proceed immediately).
        """
        with self._lock:
            now = time.monotonic()

            # Prune expired entries
            minute_cutoff = now - 60.0
            while self._minute_window and self._minute_window[0] < minute_cutoff:
                self._minute_window.popleft()

            hour_cutoff = now - 3600.0
            while self._hour_window and self._hour_window[0] < hour_cutoff:
                self._hour_window.popleft()

            # Calculate required wait
            wait = 0.0

            if len(self._minute_window) >= self._config.max_per_minute:
                oldest = self._minute_window[0]
                minute_wait = (oldest + 60.0) - now
                if minute_wait > wait:
                    wait = minute_wait

            if len(self._hour_window) >= self._config.max_per_hour:
                oldest = self._hour_window[0]
                hour_wait = (oldest + 3600.0) - now
                if hour_wait > wait:
                    wait = hour_wait

            # Clamp negative waits to zero
            if wait < 0:
                wait = 0.0

            # Record future timestamp
            request_time = now + wait
            self._minute_window.append(request_time)
            self._hour_window.append(request_time)

            return wait



"""JIT-compatible EOP interpolation and query functions.

All functions use only JAX primitives (``jnp.searchsorted``, array indexing,
``jnp.where``) and are fully compatible with ``jax.jit``, ``jax.vmap``,
and ``jax.grad``.

The ``extrapolation`` parameter is a Python string resolved at trace time,
following the same pattern as ``ForceModelConfig`` booleans.
"""

from __future__ import annotations

import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from astrojax.eop._types import EOPData, EOPExtrapolation


def _interpolate_scalar(
    eop: EOPData,
    mjd: Array,
    values: Array,
    extrapolation: EOPExtrapolation = EOPExtrapolation.HOLD,
) -> Array:
    """Linearly interpolate a single EOP field at the given MJD.

    Uses ``jnp.searchsorted`` for O(log n) lookup, then linear interpolation
    between bracketing points. Out-of-range queries are handled by the
    extrapolation mode.

    Args:
        eop: EOP dataset with sorted MJD array.
        mjd: Scalar MJD to query.
        values: The EOP field array to interpolate, shape ``(N,)``.
        extrapolation: Extrapolation mode for out-of-range queries.

    Returns:
        Interpolated scalar value.
    """
    n = eop.mjd.shape[0]

    # Binary search: idx is the insertion point (right side)
    idx = jnp.searchsorted(eop.mjd, mjd, side="right")

    # Bracket indices, clamped to valid range
    idx_lo = jnp.clip(idx - 1, 0, n - 1)
    idx_hi = jnp.clip(idx, 0, n - 1)

    mjd_lo = eop.mjd[idx_lo]
    mjd_hi = eop.mjd[idx_hi]
    val_lo = values[idx_lo]
    val_hi = values[idx_hi]

    # Linear interpolation fraction (safe division: if mjd_lo == mjd_hi, frac=0)
    dmjd = mjd_hi - mjd_lo
    frac = jnp.where(dmjd > 0.0, (mjd - mjd_lo) / dmjd, 0.0)
    interpolated = val_lo + frac * (val_hi - val_lo)

    # Extrapolation handling
    in_range = (mjd >= eop.mjd_min) & (mjd <= eop.mjd_max)

    if extrapolation == EOPExtrapolation.ZERO:
        return jnp.where(in_range, interpolated, 0.0)
    else:
        # HOLD: clamp to boundary values (already done by clip)
        return interpolated


def get_ut1_utc(
    eop: EOPData,
    mjd: ArrayLike,
    extrapolation: EOPExtrapolation = EOPExtrapolation.HOLD,
) -> Array:
    """Query UT1-UTC offset at the given MJD.

    Args:
        eop: EOP dataset.
        mjd: Modified Julian Date to query.
        extrapolation: Extrapolation mode for out-of-range queries.

    Returns:
        UT1-UTC offset [seconds].

    Examples:
        ```python
        from astrojax.eop import zero_eop, get_ut1_utc
        eop = zero_eop()
        ut1_utc = get_ut1_utc(eop, 59569.0)
        ```
    """
    mjd = jnp.asarray(mjd, dtype=eop.mjd.dtype)
    return _interpolate_scalar(eop, mjd, eop.ut1_utc, extrapolation)


def get_pm(
    eop: EOPData,
    mjd: ArrayLike,
    extrapolation: EOPExtrapolation = EOPExtrapolation.HOLD,
) -> tuple[Array, Array]:
    """Query polar motion components at the given MJD.

    Args:
        eop: EOP dataset.
        mjd: Modified Julian Date to query.
        extrapolation: Extrapolation mode for out-of-range queries.

    Returns:
        Tuple of (pm_x, pm_y) polar motion components [rad].

    Examples:
        ```python
        from astrojax.eop import zero_eop, get_pm
        eop = zero_eop()
        pm_x, pm_y = get_pm(eop, 59569.0)
        ```
    """
    mjd = jnp.asarray(mjd, dtype=eop.mjd.dtype)
    pm_x = _interpolate_scalar(eop, mjd, eop.pm_x, extrapolation)
    pm_y = _interpolate_scalar(eop, mjd, eop.pm_y, extrapolation)
    return pm_x, pm_y


def get_dxdy(
    eop: EOPData,
    mjd: ArrayLike,
    extrapolation: EOPExtrapolation = EOPExtrapolation.HOLD,
) -> tuple[Array, Array]:
    """Query celestial pole offsets at the given MJD.

    Values may be NaN if the data source does not include dX/dY for the
    requested epoch (e.g. far-future predictions).

    Args:
        eop: EOP dataset.
        mjd: Modified Julian Date to query.
        extrapolation: Extrapolation mode for out-of-range queries.

    Returns:
        Tuple of (dX, dY) celestial pole offsets [rad].

    Examples:
        ```python
        from astrojax.eop import zero_eop, get_dxdy
        eop = zero_eop()
        dx, dy = get_dxdy(eop, 59569.0)
        ```
    """
    mjd = jnp.asarray(mjd, dtype=eop.mjd.dtype)
    dx = _interpolate_scalar(eop, mjd, eop.dX, extrapolation)
    dy = _interpolate_scalar(eop, mjd, eop.dY, extrapolation)
    return dx, dy


def get_lod(
    eop: EOPData,
    mjd: ArrayLike,
    extrapolation: EOPExtrapolation = EOPExtrapolation.HOLD,
) -> Array:
    """Query length-of-day excess at the given MJD.

    The value may be NaN if the data source does not include LOD for the
    requested epoch (e.g. far-future predictions).

    Args:
        eop: EOP dataset.
        mjd: Modified Julian Date to query.
        extrapolation: Extrapolation mode for out-of-range queries.

    Returns:
        Length of day excess [seconds].

    Examples:
        ```python
        from astrojax.eop import zero_eop, get_lod
        eop = zero_eop()
        lod = get_lod(eop, 59569.0)
        ```
    """
    mjd = jnp.asarray(mjd, dtype=eop.mjd.dtype)
    return _interpolate_scalar(eop, mjd, eop.lod, extrapolation)


def get_eop(
    eop: EOPData,
    mjd: ArrayLike,
    extrapolation: EOPExtrapolation = EOPExtrapolation.HOLD,
) -> tuple[Array, Array, Array, Array, Array, Array]:
    """Query all EOP values at the given MJD.

    Convenience function that returns all six EOP parameters at once.

    Args:
        eop: EOP dataset.
        mjd: Modified Julian Date to query.
        extrapolation: Extrapolation mode for out-of-range queries.

    Returns:
        Tuple of (pm_x, pm_y, ut1_utc, lod, dX, dY).
        Units: pm_x/pm_y [rad], ut1_utc [s], lod [s], dX/dY [rad].

    Examples:
        ```python
        from astrojax.eop import zero_eop, get_eop
        eop = zero_eop()
        pm_x, pm_y, ut1_utc, lod, dx, dy = get_eop(eop, 59569.0)
        ```
    """
    mjd = jnp.asarray(mjd, dtype=eop.mjd.dtype)
    pm_x = _interpolate_scalar(eop, mjd, eop.pm_x, extrapolation)
    pm_y = _interpolate_scalar(eop, mjd, eop.pm_y, extrapolation)
    ut1_utc = _interpolate_scalar(eop, mjd, eop.ut1_utc, extrapolation)
    lod = _interpolate_scalar(eop, mjd, eop.lod, extrapolation)
    dx = _interpolate_scalar(eop, mjd, eop.dX, extrapolation)
    dy = _interpolate_scalar(eop, mjd, eop.dY, extrapolation)
    return pm_x, pm_y, ut1_utc, lod, dx, dy



"""TEME frame transformations for SGP4 output.

TEME (True Equator, Mean Equinox) is the native output frame of the SGP4/SDP4
propagator.  This module provides transformations between TEME and other
commonly used reference frames:

- **TEME -> PEF**: Rotate by GMST (removes mean Earth rotation)
- **PEF -> ITRF**: Apply polar motion
- **ITRF -> GCRF**: Full IAU 2006/2000A CIO-based transformation

All inputs and outputs use SI base units (metres, metres/second).
"""

from __future__ import annotations

import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from astrojax.attitude_representations import Rz
from astrojax.config import get_dtype
from astrojax.constants import OMEGA_EARTH
from astrojax.eop._types import EOPData
from astrojax.epoch import Epoch
from astrojax.frames.gcrf_itrf import (
    polar_motion,
    state_gcrf_to_itrf,
    state_itrf_to_gcrf,
)

# ---------------------------------------------------------------------------
# Rotation matrices
# ---------------------------------------------------------------------------


def rotation_teme_to_pef(epc: Epoch) -> Array:
    """Compute the 3x3 rotation matrix from TEME to PEF.

    Applies ``Rz(GMST)`` to rotate from the mean equinox frame to the
    pseudo Earth-fixed frame.

    Args:
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (TEME -> PEF).
    """
    return Rz(epc.gmst())


def rotation_pef_to_teme(epc: Epoch) -> Array:
    """Compute the 3x3 rotation matrix from PEF to TEME.

    Transpose of :func:`rotation_teme_to_pef`.

    Args:
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (PEF -> TEME).
    """
    return rotation_teme_to_pef(epc).T


def rotation_teme_to_itrf(eop: EOPData, epc: Epoch) -> Array:
    """Compute the 3x3 rotation matrix from TEME to ITRF.

    Combines GMST rotation and polar motion:
    ``R = PM @ Rz(GMST)``

    Args:
        eop: EOP data providing polar motion parameters.
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (TEME -> ITRF).
    """
    pm = polar_motion(eop, epc)
    r_gmst = Rz(epc.gmst())
    return pm @ r_gmst


def rotation_itrf_to_teme(eop: EOPData, epc: Epoch) -> Array:
    """Compute the 3x3 rotation matrix from ITRF to TEME.

    Transpose of :func:`rotation_teme_to_itrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (ITRF -> TEME).
    """
    return rotation_teme_to_itrf(eop, epc).T


# ---------------------------------------------------------------------------
# State vector transformations
# ---------------------------------------------------------------------------


def state_teme_to_pef(epc: Epoch, x_teme: ArrayLike) -> Array:
    """Transform a 6-element state vector from TEME to PEF.

    Velocity includes the Earth-rotation correction:
    ``v_pef = R @ v_teme - omega_earth x r_pef``

    Args:
        epc: Epoch (UTC).
        x_teme: 6-element TEME state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element PEF state ``[x, y, z, vx, vy, vz]``. Units: m, m/s.
    """
    dtype = get_dtype()
    x_teme = jnp.asarray(x_teme, dtype=dtype)

    R = rotation_teme_to_pef(epc)
    omega = jnp.array([0.0, 0.0, OMEGA_EARTH], dtype=dtype)

    r_teme = x_teme[:3]
    v_teme = x_teme[3:6]

    r_pef = R @ r_teme
    v_pef = R @ v_teme - jnp.cross(omega, r_pef)

    return jnp.concatenate([r_pef, v_pef])


def state_pef_to_teme(epc: Epoch, x_pef: ArrayLike) -> Array:
    """Transform a 6-element state vector from PEF to TEME.

    Inverse of :func:`state_teme_to_pef`.

    Args:
        epc: Epoch (UTC).
        x_pef: 6-element PEF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element TEME state ``[x, y, z, vx, vy, vz]``. Units: m, m/s.
    """
    dtype = get_dtype()
    x_pef = jnp.asarray(x_pef, dtype=dtype)

    R = rotation_teme_to_pef(epc)
    omega = jnp.array([0.0, 0.0, OMEGA_EARTH], dtype=dtype)

    r_pef = x_pef[:3]
    v_pef = x_pef[3:6]

    r_teme = R.T @ r_pef
    v_teme = R.T @ (v_pef + jnp.cross(omega, r_pef))

    return jnp.concatenate([r_teme, v_teme])


def state_teme_to_itrf(eop: EOPData, epc: Epoch, x_teme: ArrayLike) -> Array:
    """Transform a 6-element state vector from TEME to ITRF.

    Goes via PEF: applies GMST rotation and omega correction, then
    applies polar motion (position only, PM has negligible velocity effect).

    Args:
        eop: EOP data providing polar motion parameters.
        epc: Epoch (UTC).
        x_teme: 6-element TEME state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element ITRF state ``[x, y, z, vx, vy, vz]``. Units: m, m/s.
    """
    dtype = get_dtype()
    x_teme = jnp.asarray(x_teme, dtype=dtype)

    R_gmst = rotation_teme_to_pef(epc)
    pm = polar_motion(eop, epc)
    omega = jnp.array([0.0, 0.0, OMEGA_EARTH], dtype=dtype)

    r_teme = x_teme[:3]
    v_teme = x_teme[3:6]

    # TEME -> PEF
    r_pef = R_gmst @ r_teme
    v_pef = R_gmst @ v_teme - jnp.cross(omega, r_pef)

    # PEF -> ITRF (polar motion)
    r_itrf = pm @ r_pef
    v_itrf = pm @ v_pef

    return jnp.concatenate([r_itrf, v_itrf])


def state_itrf_to_teme(eop: EOPData, epc: Epoch, x_itrf: ArrayLike) -> Array:
    """Transform a 6-element state vector from ITRF to TEME.

    Inverse of :func:`state_teme_to_itrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).
        x_itrf: 6-element ITRF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element TEME state ``[x, y, z, vx, vy, vz]``. Units: m, m/s.
    """
    dtype = get_dtype()
    x_itrf = jnp.asarray(x_itrf, dtype=dtype)

    R_gmst = rotation_teme_to_pef(epc)
    pm = polar_motion(eop, epc)
    omega = jnp.array([0.0, 0.0, OMEGA_EARTH], dtype=dtype)

    r_itrf = x_itrf[:3]
    v_itrf = x_itrf[3:6]

    # ITRF -> PEF (inverse polar motion)
    r_pef = pm.T @ r_itrf
    v_pef = pm.T @ v_itrf

    # PEF -> TEME (inverse GMST rotation + omega correction)
    r_teme = R_gmst.T @ r_pef
    v_teme = R_gmst.T @ (v_pef + jnp.cross(omega, r_pef))

    return jnp.concatenate([r_teme, v_teme])


def state_teme_to_gcrf(eop: EOPData, epc: Epoch, x_teme: ArrayLike) -> Array:
    """Transform a 6-element state vector from TEME to GCRF.

    Chains TEME -> ITRF -> GCRF using the existing ITRF-GCRF
    transformation.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).
        x_teme: 6-element TEME state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element GCRF state ``[x, y, z, vx, vy, vz]``. Units: m, m/s.
    """
    x_itrf = state_teme_to_itrf(eop, epc, x_teme)
    return state_itrf_to_gcrf(eop, epc, x_itrf)


def state_gcrf_to_teme(eop: EOPData, epc: Epoch, x_gcrf: ArrayLike) -> Array:
    """Transform a 6-element state vector from GCRF to TEME.

    Chains GCRF -> ITRF -> TEME.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).
        x_gcrf: 6-element GCRF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element TEME state ``[x, y, z, vx, vy, vz]``. Units: m, m/s.
    """
    x_itrf = state_gcrf_to_itrf(eop, epc, x_gcrf)
    return state_itrf_to_teme(eop, epc, x_itrf)



"""Type definitions for Earth Orientation Parameters (EOP).

Provides the core data types for EOP storage and lookup:

- :class:`EOPData`: Immutable container holding sorted EOP arrays for
  JIT-compatible interpolation via ``jnp.searchsorted``.
- :class:`EOPExtrapolation`: Controls behavior when querying outside the
  data range.

``EOPData`` is a :class:`~typing.NamedTuple`, which JAX treats as a pytree
automatically. This means it works seamlessly with ``jax.jit``,
``jax.vmap``, and ``jax.lax`` control flow primitives.
"""

from __future__ import annotations

import enum
from typing import NamedTuple

from jax import Array


class EOPData(NamedTuple):
    """Earth Orientation Parameter data for JIT-compatible lookups.

    Stores EOP values as sorted JAX arrays, enabling O(log n) interpolation
    inside ``jax.jit`` via ``jnp.searchsorted``. Missing optional values
    (dX, dY, lod in prediction regions) are stored as NaN.

    Attributes:
        mjd: Sorted Modified Julian Dates, shape ``(N,)``.
        pm_x: Polar motion x-component [rad], shape ``(N,)``.
        pm_y: Polar motion y-component [rad], shape ``(N,)``.
        ut1_utc: UT1-UTC offset [seconds], shape ``(N,)``.
        dX: Celestial pole offset X [rad], shape ``(N,)``. NaN where missing.
        dY: Celestial pole offset Y [rad], shape ``(N,)``. NaN where missing.
        lod: Length of day excess [seconds], shape ``(N,)``. NaN where missing.
        mjd_min: Scalar, first MJD in the dataset.
        mjd_max: Scalar, last MJD in the dataset.
        mjd_last_lod: Scalar, last MJD with valid LOD data.
        mjd_last_dxdy: Scalar, last MJD with valid dX/dY data.
    """

    mjd: Array
    pm_x: Array
    pm_y: Array
    ut1_utc: Array
    dX: Array
    dY: Array
    lod: Array
    mjd_min: Array
    mjd_max: Array
    mjd_last_lod: Array
    mjd_last_dxdy: Array


class EOPExtrapolation(enum.Enum):
    """Extrapolation mode for EOP queries outside the data range.

    Resolved at trace time (Python string), not at runtime. This follows
    the same pattern as ``ForceModelConfig`` booleans.

    Attributes:
        HOLD: Clamp to the nearest boundary value.
        ZERO: Return zero for out-of-range queries.
    """

    HOLD = "hold"
    ZERO = "zero"



"""Client-side filtering engine for Celestrak query results.

Parses SpaceTrack-compatible operator strings and applies them
as filters to downloaded records.

Supported operators:

- ``>value`` -- Greater than
- ``<value`` -- Less than
- ``<>value`` -- Not equal (case-insensitive)
- ``min--max`` -- Inclusive range
- ``~~pattern`` -- Case-insensitive substring match (like)
- ``^prefix`` -- Case-insensitive prefix match (starts with)
- ``value`` -- Exact string match
"""

from __future__ import annotations

import functools
from typing import Any, Protocol


class HasGetField(Protocol):
    """Protocol for records that support field access by name."""

    def get_field(self, name: str) -> str | None: ...


def get_field(record: Any, name: str) -> str | None:
    """Get a field value from any record type.

    Supports records with a ``get_field`` method (GPRecord,
    CelestrakSATCATRecord) or arbitrary attribute access.

    Args:
        record: A record object with field access.
        name: Uppercase field name (e.g. ``"OBJECT_NAME"``).

    Returns:
        String representation of the field value, or None if unset.
    """
    if hasattr(record, "get_field"):
        return record.get_field(name)
    attr_name = name.lower()
    try:
        val = getattr(record, attr_name)
    except AttributeError:
        return None
    if val is None:
        return None
    return str(val)


def _parse_filter_value(value: str) -> tuple[str, str | tuple[str, str]]:
    """Parse a filter value string into (operator, operand).

    Args:
        value: Filter value string (e.g. ``">50"``, ``"10--20"``).

    Returns:
        A tuple of (op_type, operand) where op_type is one of:
        ``"gt"``, ``"lt"``, ``"ne"``, ``"range"``, ``"like"``,
        ``"startswith"``, ``"exact"``.
        For ``"range"``, operand is a tuple ``(min, max)``.
    """
    if value.startswith("<>"):
        return ("ne", value[2:])
    if value.startswith("~~"):
        return ("like", value[2:])
    if value.startswith(">"):
        return ("gt", value[1:])
    if value.startswith("<"):
        return ("lt", value[1:])
    if value.startswith("^"):
        return ("startswith", value[1:])
    dash_pos = value.find("--")
    if dash_pos >= 0:
        return ("range", (value[:dash_pos], value[dash_pos + 2:]))
    return ("exact", value)


def _compare_values(a: str, b: str) -> int | None:
    """Compare two string values, attempting numeric comparison first.

    Args:
        a: First value.
        b: Second value.

    Returns:
        -1 if a < b, 0 if a == b, 1 if a > b, None if incomparable.
    """
    try:
        a_num = float(a)
        b_num = float(b)
        if a_num < b_num:
            return -1
        elif a_num > b_num:
            return 1
        else:
            return 0
    except (ValueError, TypeError):
        pass
    if a < b:
        return -1
    elif a > b:
        return 1
    else:
        return 0


def _matches_filter(record: Any, field: str, value: str) -> bool:
    """Check if a single filter matches a record.

    Args:
        record: A record object with field access.
        field: Uppercase field name.
        value: Filter value string with optional operator prefix.

    Returns:
        True if the record matches the filter.
    """
    field_value = get_field(record, field)
    if field_value is None:
        return False

    op_type, operand = _parse_filter_value(value)

    if op_type == "gt":
        return _compare_values(field_value, operand) == 1
    elif op_type == "lt":
        return _compare_values(field_value, operand) == -1
    elif op_type == "ne":
        return field_value.lower() != operand.lower()
    elif op_type == "range":
        min_val, max_val = operand
        cmp_min = _compare_values(field_value, min_val)
        cmp_max = _compare_values(field_value, max_val)
        return (
            cmp_min is not None
            and cmp_max is not None
            and cmp_min >= 0
            and cmp_max <= 0
        )
    elif op_type == "like":
        return operand.lower() in field_value.lower()
    elif op_type == "startswith":
        return field_value.lower().startswith(operand.lower())
    else:  # exact
        return field_value == operand


def apply_filters(
    records: list[Any], filters: list[tuple[str, str]]
) -> list[Any]:
    """Apply client-side filters to a list of records.

    Records must match ALL filters (AND logic).

    Args:
        records: List of records to filter.
        filters: List of (field, value) tuples.

    Returns:
        Filtered list of records.
    """
    if not filters:
        return records
    return [
        r
        for r in records
        if all(_matches_filter(r, field, value) for field, value in filters)
    ]


def apply_order_by(
    records: list[Any], order_by: list[tuple[str, bool]]
) -> None:
    """Apply client-side ordering to a list of records (in-place).

    Args:
        records: List of records to sort.
        order_by: List of (field, ascending) tuples.
    """
    if not order_by:
        return

    def compare(a: Any, b: Any) -> int:
        for field, ascending in order_by:
            a_val = get_field(a, field)
            b_val = get_field(b, field)
            if a_val is not None and b_val is not None:
                cmp = _compare_values(a_val, b_val)
                if cmp is None:
                    cmp = 0
            elif a_val is not None:
                cmp = -1
            elif b_val is not None:
                cmp = 1
            else:
                cmp = 0
            if not ascending:
                cmp = -cmp
            if cmp != 0:
                return cmp
        return 0

    records.sort(key=functools.cmp_to_key(compare))


def apply_limit(records: list[Any], limit: int | None) -> list[Any]:
    """Apply a client-side limit to truncate results.

    Args:
        records: List of records.
        limit: Maximum number of records, or None for no limit.

    Returns:
        Truncated list of records.
    """
    if limit is None:
        return records
    return records[:limit]



"""GCRF-ITRF frame transformations using the IAU 2006/2000A CIO-based model.

Provides high-fidelity rotation matrices and state-vector transformations
between the Geocentric Celestial Reference Frame (GCRF) and the International
Terrestrial Reference Frame (ITRF), implementing:

- **Bias-precession-nutation** (GCRF -> CIRS) via IAU 2006/2000A
- **Earth rotation** (CIRS -> TIRS) via IAU 2000 Earth Rotation Angle
- **Polar motion** (TIRS -> ITRF)

All functions take an explicit ``eop: EOPData`` parameter (no global state).

ECI/ECEF aliases are provided for backward compatibility:
:func:`rotation_eci_to_ecef` == :func:`rotation_gcrf_to_itrf`, etc.

All inputs and outputs use SI base units (metres, metres/second).

Uses routines and computations derived from software provided by SOFA
under license. Does not itself constitute software provided by and/or
endorsed by SOFA.
"""

from __future__ import annotations

import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from astrojax.attitude_representations import Rz
from astrojax.config import get_dtype
from astrojax.constants import OMEGA_EARTH
from astrojax.eop._lookup import get_dxdy, get_pm, get_ut1_utc
from astrojax.eop._types import EOPData
from astrojax.epoch import Epoch
from astrojax.sofa import (
    MJD_ZERO,
    c2ixys,
    era00,
    pom00,
    sp00,
    xys06a,
)
from astrojax.time import TT_TAI, leap_seconds_tai_utc


def _mjd_tt(eop: EOPData, epc: Epoch) -> Array:
    """Compute MJD in TT time system from a UTC epoch.

    TT = UTC + (TAI-UTC) + TT_TAI, where TAI-UTC is the leap second count.

    Args:
        eop: EOP data (unused, but kept for API consistency).
        epc: Epoch (UTC).

    Returns:
        MJD in TT.
    """
    mjd_utc = epc.mjd()
    tai_utc = leap_seconds_tai_utc(mjd_utc)
    return mjd_utc + (tai_utc + TT_TAI) / 86400.0


def _mjd_ut1(eop: EOPData, epc: Epoch) -> Array:
    """Compute MJD in UT1 time system from a UTC epoch.

    UT1 = UTC + (UT1-UTC), where UT1-UTC comes from EOP data.

    Args:
        eop: EOP data providing UT1-UTC offset.
        epc: Epoch (UTC).

    Returns:
        MJD in UT1.
    """
    mjd_utc = epc.mjd()
    ut1_utc = get_ut1_utc(eop, mjd_utc)
    return mjd_utc + ut1_utc / 86400.0


# ---------------------------------------------------------------------------
# Core rotation components
# ---------------------------------------------------------------------------


def bias_precession_nutation(eop: EOPData, epc: Epoch) -> Array:
    """Compute the bias-precession-nutation matrix (GCRF -> CIRS).

    Uses the IAU 2006/2000A CIO-based model with dX/dY corrections from
    EOP data applied to the CIP coordinates.

    Args:
        eop: EOP data providing dX, dY celestial pole offsets.
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (GCRF -> CIRS).
    """
    mjd_tt = _mjd_tt(eop, epc)

    # CIP X, Y and CIO locator s from IAU 2006/2000A model
    x, y, s = xys06a(MJD_ZERO, mjd_tt)

    # Apply celestial pole offsets from EOP data (at UTC)
    mjd_utc = epc.mjd()
    dx, dy = get_dxdy(eop, mjd_utc)
    x = x + dx
    y = y + dy

    # Form celestial-to-intermediate matrix
    return c2ixys(x, y, s)


def earth_rotation_angle(eop: EOPData, epc: Epoch) -> Array:
    """Compute the Earth rotation matrix (CIRS -> TIRS).

    Uses the IAU 2000 Earth Rotation Angle with UT1 time from EOP data.

    Args:
        eop: EOP data providing UT1-UTC offset.
        epc: Epoch (UTC).

    Returns:
        3x3 Earth rotation matrix.
    """
    mjd_ut1 = _mjd_ut1(eop, epc)
    era = era00(MJD_ZERO, mjd_ut1)
    return Rz(era)


def polar_motion(eop: EOPData, epc: Epoch) -> Array:
    """Compute the polar motion matrix (TIRS -> ITRF).

    Uses polar motion parameters from EOP data and the TIO locator s'.

    Args:
        eop: EOP data providing pm_x, pm_y polar motion parameters.
        epc: Epoch (UTC).

    Returns:
        3x3 polar motion matrix.
    """
    mjd_tt = _mjd_tt(eop, epc)
    mjd_utc = epc.mjd()

    # Polar motion parameters from EOP (queried at UTC)
    pm_x, pm_y = get_pm(eop, mjd_utc)

    # TIO locator s' (function of TT)
    sp_val = sp00(MJD_ZERO, mjd_tt)

    return pom00(pm_x, pm_y, sp_val)


# ---------------------------------------------------------------------------
# Combined transformations (GCRF/ITRF naming)
# ---------------------------------------------------------------------------


def rotation_gcrf_to_itrf(eop: EOPData, epc: Epoch) -> Array:
    """Compute the full 3x3 rotation matrix from GCRF to ITRF.

    Combines polar motion, Earth rotation, and bias-precession-nutation:
    ``R = PM @ ER @ BPN``

    Args:
        eop: EOP data.
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (GCRF -> ITRF).

    Examples:
        ```python
        from astrojax import Epoch
        from astrojax.eop import zero_eop
        from astrojax.frames import rotation_gcrf_to_itrf
        eop = zero_eop()
        epc = Epoch(2024, 1, 1)
        R = rotation_gcrf_to_itrf(eop, epc)
        R.shape
        ```
    """
    bpn = bias_precession_nutation(eop, epc)
    er = earth_rotation_angle(eop, epc)
    pm = polar_motion(eop, epc)
    return pm @ er @ bpn


def rotation_itrf_to_gcrf(eop: EOPData, epc: Epoch) -> Array:
    """Compute the full 3x3 rotation matrix from ITRF to GCRF.

    This is the transpose of :func:`rotation_gcrf_to_itrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (ITRF -> GCRF).
    """
    return rotation_gcrf_to_itrf(eop, epc).T


def state_gcrf_to_itrf(eop: EOPData, epc: Epoch, x_gcrf: ArrayLike) -> Array:
    """Transform a 6-element state vector from GCRF to ITRF.

    Position is rotated directly. Velocity includes the correction for
    Earth's rotation (Coriolis effect):

    .. math::

        \\mathbf{r}_{\\text{ITRF}} &= PM \\cdot ER \\cdot BPN \\cdot
            \\mathbf{r}_{\\text{GCRF}} \\\\
        \\mathbf{v}_{\\text{ITRF}} &= PM \\cdot \\left(
            ER \\cdot BPN \\cdot \\mathbf{v}_{\\text{GCRF}}
            - \\boldsymbol{\\omega} \\times (ER \\cdot BPN \\cdot
            \\mathbf{r}_{\\text{GCRF}}) \\right)

    Args:
        eop: EOP data.
        epc: Epoch (UTC).
        x_gcrf: 6-element GCRF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element ITRF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.
    """
    dtype = get_dtype()
    x_gcrf = jnp.asarray(x_gcrf, dtype=dtype)

    bpn = bias_precession_nutation(eop, epc)
    er = earth_rotation_angle(eop, epc)
    pm = polar_motion(eop, epc)

    omega = jnp.array([0.0, 0.0, OMEGA_EARTH], dtype=dtype)

    r_gcrf = x_gcrf[:3]
    v_gcrf = x_gcrf[3:6]

    # Position: PM @ ER @ BPN @ r
    r_cirs = bpn @ r_gcrf
    r_tirs = er @ r_cirs
    r_itrf = pm @ r_tirs

    # Velocity: PM @ (ER @ BPN @ v - omega x (ER @ BPN @ r))
    v_cirs = bpn @ v_gcrf
    v_tirs = er @ v_cirs - jnp.cross(omega, r_tirs)
    v_itrf = pm @ v_tirs

    return jnp.concatenate([r_itrf, v_itrf])


def state_itrf_to_gcrf(eop: EOPData, epc: Epoch, x_itrf: ArrayLike) -> Array:
    """Transform a 6-element state vector from ITRF to GCRF.

    Applies the inverse of :func:`state_gcrf_to_itrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).
        x_itrf: 6-element ITRF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element GCRF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.
    """
    dtype = get_dtype()
    x_itrf = jnp.asarray(x_itrf, dtype=dtype)

    bpn = bias_precession_nutation(eop, epc)
    er = earth_rotation_angle(eop, epc)
    pm = polar_motion(eop, epc)

    omega = jnp.array([0.0, 0.0, OMEGA_EARTH], dtype=dtype)

    r_itrf = x_itrf[:3]
    v_itrf = x_itrf[3:6]

    # Inverse: r_tirs = PM^T @ r_itrf, etc.
    r_tirs = pm.T @ r_itrf
    v_tirs = pm.T @ v_itrf

    r_cirs = er.T @ r_tirs
    v_cirs = er.T @ (v_tirs + jnp.cross(omega, r_tirs))

    r_gcrf = bpn.T @ r_cirs
    v_gcrf = bpn.T @ v_cirs

    return jnp.concatenate([r_gcrf, v_gcrf])


# ---------------------------------------------------------------------------
# ECI/ECEF aliases (backward compatibility)
# ---------------------------------------------------------------------------


def earth_rotation(eop: EOPData, epc: Epoch) -> Array:
    """Compute the Earth rotation matrix at the given epoch.

    This is the full IAU 2006/2000A GCRF -> ITRF rotation combining
    bias-precession-nutation, Earth rotation angle, and polar motion.
    Equivalent to :func:`rotation_gcrf_to_itrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).

    Returns:
        3x3 Earth rotation matrix (GCRF -> ITRF).
    """
    return rotation_gcrf_to_itrf(eop, epc)


def rotation_eci_to_ecef(eop: EOPData, epc: Epoch) -> Array:
    """Compute the 3x3 rotation matrix from ECI (GCRF) to ECEF (ITRF).

    Alias for :func:`rotation_gcrf_to_itrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (ECI -> ECEF).
    """
    return rotation_gcrf_to_itrf(eop, epc)


def rotation_ecef_to_eci(eop: EOPData, epc: Epoch) -> Array:
    """Compute the 3x3 rotation matrix from ECEF (ITRF) to ECI (GCRF).

    Alias for :func:`rotation_itrf_to_gcrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).

    Returns:
        3x3 rotation matrix (ECEF -> ECI).
    """
    return rotation_itrf_to_gcrf(eop, epc)


def state_eci_to_ecef(eop: EOPData, epc: Epoch, x_eci: ArrayLike) -> Array:
    """Transform a 6-element state vector from ECI (GCRF) to ECEF (ITRF).

    Alias for :func:`state_gcrf_to_itrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).
        x_eci: 6-element ECI state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element ECEF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.
    """
    return state_gcrf_to_itrf(eop, epc, x_eci)


def state_ecef_to_eci(eop: EOPData, epc: Epoch, x_ecef: ArrayLike) -> Array:
    """Transform a 6-element state vector from ECEF (ITRF) to ECI (GCRF).

    Alias for :func:`state_itrf_to_gcrf`.

    Args:
        eop: EOP data.
        epc: Epoch (UTC).
        x_ecef: 6-element ECEF state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.

    Returns:
        6-element ECI state ``[x, y, z, vx, vy, vz]``.
            Units: m, m/s.
    """
    return state_itrf_to_gcrf(eop, epc, x_ecef)
